'''Implementation of cbeub: Capcom Beat 'em Up Bundle'''
import logging
import os

from gex.lib.archive import arc
from gex.lib.utils.vendor import capcom
from gex.lib.utils.blob import transforms
from gex.lib.tasks import helpers

from gex.lib.tasks.basetask import BaseTask

logger = logging.getLogger('gextoolbox')

class CBEUBTask(BaseTask):
    '''Implements cbeub: Capcom Beat 'em Up Bundle'''
    _task_name = "cbeub"
    _title = "Capcom Beat 'em Up Bundle"
    _details_markdown = '''
The notes I found were at https://web.archive.org/web/20220213232104/http://blog.livedoor.jp/scrap_a/archives/23114141.html.
This was a Japanese set of shell scripts and odd generic operation executables. There is some weird encoding here too.
This script will extract and prep the ROMs. Some per-rom errata are in the notes below.
'''

    _default_input_folder = helpers.gen_steam_app_default_folder("CBEUB")
    _input_folder_desc = "CBEUB Steam folder"

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        # for each output file entry
        for out_file_entry in self._metadata['out']['files']:
            pkg_name = out_file_entry['in_file']
            # Check the status of it
            if out_file_entry['status'] == 'no-rom':
                logger.info(f"Skipping {pkg_name} - cannot extract...")
            else:
                logger.info(f"Extracting {pkg_name}...")

                # read the matching input file
                in_file_entry = self._metadata['in']['files'][pkg_name]
                loaded_file = self.read_datafile(in_dir, in_file_entry)

                # extract the input file
                arc_contents = arc.extract(loaded_file['contents'])

                # Get the bin entry
                merged_rom_contents = None
                for _, arc_content in arc_contents.items():
                    if arc_content['path'].startswith('bin'):
                        merged_rom_contents = arc_content['contents']

                # run the handler
                handler_func = self.find_handler_func(pkg_name)
                if merged_rom_contents is not None and handler_func is not None:
                    output_contents = handler_func(merged_rom_contents)

                    _ = self.verify_out_file(out_file_entry['filename'], output_contents)

                    with open(os.path.join(out_dir, out_file_entry['filename']), "wb") as out_file:
                        out_file.write(output_contents)
                elif merged_rom_contents is None:
                    logger.warning("Could not find merged rom data in arc.")
                elif handler_func is None:
                    logger.warning("Could not find matching handler function.")
        logger.info("Processing complete.")

    def _deshuffle_gfx_common(self, start, length, filenames, num_deinterleave_split, do_split):
        def gfx(contents):
            # Cut out the section
            contents = contents[start:start + length]

            # This is weird... it's a bit shuffle, not byte-level and not a normal interleave
            bit_order = [
                7, 3, 15, 11, 23, 19, 31, 27,
                6, 2, 14, 10, 22, 18, 30, 26,
                5, 1, 13, 9, 21, 17, 29, 25,
                4, 0, 12, 8, 20, 16, 28, 24,
                39, 35, 47, 43, 55, 51, 63, 59,
                38, 34, 46, 42, 54, 50, 62, 58,
                37, 33, 45, 41, 53, 49, 61, 57,
                36, 32, 44, 40, 52, 48, 60, 56
            ]
            chunks = transforms.split_bit_shuffle(
                contents, word_size_bytes=8, bit_order=bit_order, num_ways=num_deinterleave_split)

            # Split it
            if do_split:
                new_chunks = []
                for oldchunk in chunks:
                    new_chunks.extend(
                        transforms.equal_split(oldchunk, num_chunks=2))
                chunks = new_chunks

            return dict(zip(filenames, chunks))
        return gfx

    def _audio_common(self, start, filenames):
        def audio(contents):
            chunks = []

            # Add the audio CPU
            chunks.append(contents[start:start + 0x8000] +
                          contents[start + 0x10000:start + 0x18000])

            # Add the qsound
            qsound_start = start + 0x18000
            qsound_contents = contents[qsound_start:qsound_start + 0x40000]
            chunks.extend(transforms.equal_split(
                qsound_contents, num_chunks=2))

            return dict(zip(filenames, chunks))
        return audio

    ################################################################################
    # Final Fight                                                                  #
    ################################################################################

    def _handle_ffight(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "ff_42.11h",
            "ffe_43.12h",
            "ff_36.11f",
            "ff_37.12f",
            "ff-32m.8h"
        ]

        def maincpu(contents):
            chunk_5 = contents[0x080040:0x100040]
            contents = contents[0x40:0x080040]
            chunks = transforms.deinterleave(contents, num_ways=2, word_size=1)

            new_chunks = []
            for oldchunk in chunks:
                new_chunks.extend(
                    transforms.equal_split(oldchunk, num_chunks=2))
            chunks = new_chunks

            # Add 5th non-interleaved chunk
            chunks.append(chunk_5)

            return dict(zip(maincpu_filenames, chunks))
        func_map['maincpu'] = maincpu

        gfx_filenames = [
            "ff-5m.7a",
            "ff-7m.9a",
            "ff-1m.3a",
            "ff-3m.5a"
        ]
        func_map['gfx'] = self._deshuffle_gfx_common(
            0x400040, 0x200000, gfx_filenames, 4, False)

        audio_filenames = [
            'ff_09.12b',
            'ff_18.11c',
            'ff_19.12c'
        ]
        func_map['audio'] = self._audio_common(0x600040, audio_filenames)

        ph_files = {
            'buf1': 0x117,
            'ioa1': 0x117,
            'prg1': 0x117,
            'rom1': 0x117,
            'sou1': 0x117,
            's224b.1a': 0x117,
            'iob1.11e': 0x117
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_ffightj(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "ff42.bin",
            "ff43.bin",
            "ffj_40.10h",
            "ffj_41.11h",
            "ff36.bin",
            "ff37.bin",
            "ffj_34.10f",
            "ffj_35.11f"
        ]

        def maincpu(contents):
            contents = contents[0x40:0x100040]
            chunks = transforms.deinterleave(contents, num_ways=2, word_size=1)
            new_chunks = []
            for oldchunk in chunks:
                new_chunks.extend(
                    transforms.equal_split(oldchunk, num_chunks=4))
            chunks = new_chunks
            return dict(zip(maincpu_filenames, chunks))
        func_map['maincpu'] = maincpu

        gfx_filenames = [
            "ffj_09.4b",
            "ffj_10.5b",
            "ffj_01.4a",
            "ffj_02.5a",
            "ffj_13.9b",
            "ffj_14.10b",
            "ffj_05.9a",
            "ffj_06.10a",
            "ffj_24.5e",
            "ffj_25.7e",
            "ffj_17.5c",
            "ffj_18.7c",
            "ffj_38.8h",
            "ffj_39.9h",
            "ffj_32.8f",
            "ffj_33.9f"
        ]
        func_map['gfx'] = self._deshuffle_gfx_common(
            0x400040, 0x200000, gfx_filenames, 8, True)

        audio_filenames = [
            'ff_23.bin',
            'ffj_30.bin',
            'ffj_31.bin'
        ]
        func_map['audio'] = self._audio_common(0x600040, audio_filenames)

        ph_files = {
            'buf1': 0x117,
            'ioa1': 0x117,
            'prg1': 0x117,
            'rom1': 0x117,
            'sou1': 0x117,
            's222b.1a': 0x117,
            'lwio.12c': 0x117
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # The King of Dragons                                                          #
    ################################################################################

    def _handle_kod(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "kde_37a.11f",
            "kde_38a.12f",
            "kd_35.9f",
            "kd_36a.10f",
            "kde_30a.11e",
            "kde_31a.12e",
            "kd_28.9e",
            "kd_29.10e"
        ]

        def maincpu(contents):
            contents = contents[0x40:0x100040]
            chunks = transforms.deinterleave(contents, num_ways=2, word_size=1)

            new_chunks = []
            for oldchunk in chunks:
                new_chunks.extend(
                    transforms.equal_split(oldchunk, num_chunks=4))
            chunks = new_chunks

            return dict(zip(maincpu_filenames, chunks))
        func_map['maincpu'] = maincpu

        gfx_filenames = [
            "kd-5m.4a",
            "kd-6m.4c",
            "kd-7m.6a",
            "kd-8m.6c",
            "kd-1m.3a",
            "kd-2m.3c",
            "kd-3m.5a",
            "kd-4m.5c"
        ]
        func_map['gfx'] = self._deshuffle_gfx_common(
            0x400040, 0x400000, gfx_filenames, 4, True)

        audio_filenames = [
            'kd_9.12a',
            'kd_18.11c',
            'kd_19.12c'
        ]
        func_map['audio'] = self._audio_common(0x800040, audio_filenames)

        ph_files = {
            'buf1': 0x117,
            'ioa1': 0x117,
            'prg1': 0x117,
            'rom1': 0x117,
            'sou1': 0x117,
            'kd29b.1a': 0x117,
            'iob1.11d': 0x117,
            'ioc1.ic7': 0x104,
            'c632.ic1': 0x117
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_kodj(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "kdj_37a.11f",
            "kdj_38a.12f",
            "kdj_30a.11e",
            "kdj_31a.12e",
            "kd_33.6f"
        ]

        def maincpu(contents):
            chunk_5 = contents[0x080040:0x100040]
            contents = contents[0x40:0x080040]
            chunks = transforms.deinterleave(contents, num_ways=2, word_size=1)
            new_chunks = []
            for oldchunk in chunks:
                new_chunks.extend(
                    transforms.equal_split(oldchunk, num_chunks=2))
            chunks = new_chunks

            # Add 5th non-interleaved chunk
            chunks.append(chunk_5)

            return dict(zip(maincpu_filenames, chunks))
        func_map['maincpu'] = maincpu

        gfx_filenames = [
            "kd_06.8a",
            "kd_15.8c",
            "kd_08.10a",
            "kd_17.10c",
            "kd_05.7a",
            "kd_14.7c",
            "kd_07.9a",
            "kd_16.9c"
        ]
        func_map['gfx'] = self._deshuffle_gfx_common(
            0x400040, 0x400000, gfx_filenames, 4, True)

        audio_filenames = [
            'kd_09.12a',
            'kd_18.11c',
            'kd_19.12c'
        ]
        func_map['audio'] = self._audio_common(0x800040, audio_filenames)

        ph_files = {
            'buf1': 0x117,
            'ioa1': 0x117,
            'prg1': 0x117,
            'rom1': 0x117,
            'sou1': 0x117,
            'kd29b.1a': 0x117,
            'iob1.11d': 0x117,
            'ioc1.ic7': 0x104,
            'c632.ic1': 0x117
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Captain Commando                                                             #
    ################################################################################

    def _handle_captcomm(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "cce_23f.8f",
            "cc_28f.9f",
            "cc_22f.7f",
            "cc_24f.9e"
        ]

        def maincpu(contents):
            # Only the last 2 128k chunks actually need deinterleaved...
            maincpu_area = contents[0x40:0x140040]
            deint_chunks = transforms.deinterleave(
                maincpu_area[0x100000:0x140000], num_ways=2, word_size=1)
            chunks = []
            chunks.append(maincpu_area[0x0:0x80000])
            chunks.append(deint_chunks[0])
            chunks.append(maincpu_area[0x80000:0x100000])
            chunks.append(deint_chunks[1])
            return dict(zip(maincpu_filenames, chunks))
        func_map['maincpu'] = maincpu

        gfx_filenames = [
            "cc-5m.3a",
            "cc-6m.7a",
            "cc-7m.5a",
            "cc-8m.9a",
            "cc-1m.4a",
            "cc-2m.8a",
            "cc-3m.6a",
            "cc-4m.10a"
        ]
        func_map['gfx'] = self._deshuffle_gfx_common(
            0x400040, 0x400000, gfx_filenames, 4, True)

        audio_filenames = [
            'cc_09.11a',
            'cc_18.11c',
            'cc_19.12c'
        ]
        func_map['audio'] = self._audio_common(0x800040, audio_filenames)

        ph_files = {
            'buf1': 0x117,
            'ioa1': 0x117,
            'prg1': 0x117,
            'rom1': 0x117,
            'sou1': 0x117,
            'cc63b.1a': 0x117,
            'iob1.12d': 0x117,
            'ccprg1.11d': 0x117,
            'ioc1.ic7': 0x104,
            'c632b.ic1': 0x117
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_captcommj(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "ccj_23f.8f",
            "ccj_28f.9f",
            "ccj_22f.7f",
            "ccj_24f.9e"
        ]

        def maincpu(contents):
            # Only the last 2 128k chunks actually need deinterleaved...
            maincpu_area = contents[0x40:0x140040]
            deint_chunks = transforms.deinterleave(
                maincpu_area[0x100000:0x140000], num_ways=2, word_size=1)

            chunks = []
            chunks.append(maincpu_area[0x0:0x80000])
            chunks.append(deint_chunks[0])
            chunks.append(maincpu_area[0x80000:0x100000])
            chunks.append(deint_chunks[1])

            return dict(zip(maincpu_filenames, chunks))
        func_map['maincpu'] = maincpu

        gfx_filenames = [
            "cc_01.3a",
            "cc_05.7a",
            "cc_02.4a",
            "cc_06.8a",
            "cc_03.5a",
            "cc_07.9a",
            "cc_04.6a",
            "cc_08.10a"
        ]
        func_map['gfx'] = self._deshuffle_gfx_common(
            0x400040, 0x400000, gfx_filenames, 4, True)

        audio_filenames = [
            'ccj_09.12a',
            'ccj_18.11c',
            'ccj_19.12c'
        ]
        func_map['audio'] = self._audio_common(0x800040, audio_filenames)

        ph_files = {
            'buf1': 0x117,
            'ioa1': 0x117,
            'prg1': 0x117,
            'rom1': 0x117,
            'sou1': 0x117,
            'cc63b.1a': 0x117,
            'iob1.12d': 0x117,
            'ccprg1.11d': 0x117,
            'ioc1.ic7': 0x104,
            'c632.ic1': 0x117
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Knights of the Round                                                         #
    ################################################################################

    def _handle_knights(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "kr_23e.8f",
            "kr_22.7f"
        ]

        def maincpu(contents):
            # Only the last 2 128k chunks actually need deinterleaved...
            maincpu_area = contents[0x40:0x100040]
            chunks = transforms.equal_split(maincpu_area, num_chunks=2)

            return dict(zip(maincpu_filenames, chunks))
        func_map['maincpu'] = maincpu

        gfx_filenames = [
            "kr-5m.3a",
            "kr-6m.7a",
            "kr-7m.5a",
            "kr-8m.9a",
            "kr-1m.4a",
            "kr-2m.8a",
            "kr-3m.6a",
            "kr-4m.10a"
        ]
        func_map['gfx'] = self._deshuffle_gfx_common(
            0x400040, 0x400000, gfx_filenames, 4, True)

        audio_filenames = [
            'kr_09.11a',
            'kr_18.11c',
            'kr_19.12c'
        ]
        func_map['audio'] = self._audio_common(0x800040, audio_filenames)

        ph_files = {
            'buf1': 0x117,
            'ioa1': 0x117,
            'prg1': 0x117,
            'rom1': 0x117,
            'sou1': 0x117,
            'kr63b.1a': 0x117,
            'iob1.12d': 0x117,
            'bprg1.11d': 0x117,
            'ioc1.ic7': 0x104,
            'c632.ic1': 0x117
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_knightsj(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "kr_23j.8f",
            "kr_22.7f"
        ]

        def maincpu(contents):
            # Only the last 2 128k chunks actually need deinterleaved...
            maincpu_area = contents[0x40:0x100040]
            chunks = transforms.equal_split(maincpu_area, num_chunks=2)

            return dict(zip(maincpu_filenames, chunks))
        func_map['maincpu'] = maincpu

        gfx_filenames = [
            "kr_01.3a",
            "kr_05.7a",
            "kr_02.4a",
            "kr_06.8a",
            "kr_03.5a",
            "kr_07.9a",
            "kr_04.6a",
            "kr_08.10a"
        ]
        func_map['gfx'] = self._deshuffle_gfx_common(
            0x400040, 0x400000, gfx_filenames, 4, True)

        audio_filenames = [
            'kr_09.12a',
            'kr_18.11c',
            'kr_19.12c'
        ]
        func_map['audio'] = self._audio_common(0x800040, audio_filenames)

        ph_files = {
            'buf1': 0x117,
            'ioa1': 0x117,
            'prg1': 0x117,
            'rom1': 0x117,
            'sou1': 0x117,
            'kr63b.1a': 0x117,
            'iob1.12d': 0x117,
            'bprg1.11d': 0x117,
            'ioc1.ic7': 0x104,
            'c632.ic1': 0x117
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Warriors of Fate                                                             #
    ################################################################################

    def _wof_audio(self, filenames):
        def audio(contents):
            start = 0x800040
            chunks = []

            # Add the audio CPU
            audiocpu_content = contents[start:start + 0x8000] + \
                contents[start + 0x10000:start + 0x28000]
            chunks.append(audiocpu_content)

            # Add the qsound
            qsound_start = start + 0x50000
            qsound_contents = contents[qsound_start:qsound_start + 0x200000]
            chunks.extend(transforms.equal_split(
                qsound_contents, num_chunks=4))

            return dict(zip(filenames, chunks))
        return audio

    def _handle_wof(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "tk2e_23c.8f",
            "tk2e_22c.7f"
        ]

        def maincpu(contents):
            maincpu_area = contents[0x40:0x100040]
            chunks = transforms.equal_split(maincpu_area, num_chunks=2)

            return dict(zip(maincpu_filenames, chunks))
        func_map['maincpu'] = maincpu

        gfx_filenames = [
            "tk2-1m.3a",
            "tk2-5m.7a",
            "tk2-3m.5a",
            "tk2-7m.9a",
            "tk2-2m.4a",
            "tk2-6m.8a",
            "tk2-4m.6a",
            "tk2-8m.10a"
        ]

        func_map['gfx'] = self._deshuffle_gfx_common(
            0x400040, 0x400000, gfx_filenames, 4, True)

        audio_filenames = [
            'tk2_qa.5k',
            'tk2-q1.1k',
            'tk2-q2.2k',
            'tk2-q3.3k',
            'tk2-q4.4k'
        ]
        func_map['audio'] = self._wof_audio(audio_filenames)

        ph_files = {
            'buf1': 0x117,
            'ioa1': 0x117,
            'prg2': 0x117,
            'rom1': 0x117,
            'tk263b.1a': 0x117,
            'iob1.12d': 0x117,
            'bprg1.11d': 0x117,
            'ioc1.ic1': 0x104,
            'd7l1.7l': 0x117,
            'd8l1.8l': 0x117,
            'd9k1.9k': 0x117,
            'd10f1.10f': 0x117
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_wofj(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "tk2j_23c.8f",
            "tk2j_22c.7f"
        ]

        def maincpu(contents):
            maincpu_area = contents[0x40:0x100040]
            chunks = transforms.equal_split(maincpu_area, num_chunks=2)

            return dict(zip(maincpu_filenames, chunks))
        func_map['maincpu'] = maincpu

        gfx_filenames = [
            "tk2_01.3a",
            "tk2_05.7a",
            "tk2_02.4a",
            "tk2_06.8a",
            "tk2_03.5a",
            "tk2_07.9a",
            "tk2_04.6a",
            "tk2_08.10a"
        ]

        func_map['gfx'] = self._deshuffle_gfx_common(
            0x400040, 0x400000, gfx_filenames, 4, True)

        audio_filenames = [
            'tk2_qa.5k',
            'tk2-q1.1k',
            'tk2-q2.2k',
            'tk2-q3.3k',
            'tk2-q4.4k'
        ]
        func_map['audio'] = self._wof_audio(audio_filenames)

        ph_files = {
            'buf1': 0x117,
            'ioa1': 0x117,
            'prg1': 0x117,
            'sou1': 0x117,
            'rom1': 0x117,
            'tk263b.1a': 0x117,
            'iob1.12d': 0x117,
            'bprg1.11d': 0x117,
            'ioc1.ic1': 0x104,
            'd7l1.7l': 0x117,
            'd8l1.8l': 0x117,
            'd9k1.9k': 0x117,
            'd10f1.10f': 0x117
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Armored Warriors                                                             #
    ################################################################################

    def _armwar_gfx(self, contents):
        contents = contents[0x0800040:0x1C00040]

        contents = capcom.common_gfx_deshuffle(contents)

        chunks = transforms.equal_split(contents, num_chunks=20)

        # Interleave each pair of chunks
        new_chunks = []
        for oddchunk, evenchunk in zip(chunks[0::2], chunks[1::2]):
            new_chunks.append(transforms.interleave(
                [oddchunk, evenchunk], word_size=8))
        chunks = new_chunks

        contents = transforms.merge(chunks)

        # Deinterleave the chunks into our files
        new_chunks = []
        chunks = transforms.deinterleave(contents, num_ways=4, word_size=2)
        for chunk in chunks:
            new_chunks.extend(transforms.custom_split(
                chunk, [0x400000, 0x100000]))
        chunks = new_chunks
        filenames = [
            'pwg.13m',
            'pwg.14m',
            'pwg.15m',
            'pwg.16m',
            'pwg.17m',
            'pwg.18m',
            'pwg.19m',
            'pwg.20m'
        ]
        return dict(zip(filenames, chunks))

    def _armwar_audio(self, contents):
        chunks = []
        chunks.append(contents[0x1C00040:0x1C08040] +
                      contents[0x1C10040:0x1C28040])
        chunks.append(contents[0x1C28040:0x1C48040])
        filenames = [
            'pwg.01',
            'pwg.02'
        ]
        return dict(zip(filenames, chunks))

    def _armwar_qsound(self, contents):
        contents = contents[0x1C50040:0x2050040]
        chunks = transforms.equal_split(contents, num_chunks=2)
        chunks = transforms.swap_endian_all(chunks)
        filenames = [
            'pwg.11m',
            'pwg.12m'
        ]
        return dict(zip(filenames, chunks))

    def _handle_armwar(self, merged_contents):

        def maincpu(contents):
            contents = contents[0x40:0x400040]
            chunks = transforms.equal_split(contents, num_chunks=8)
            filenames = [
                "pwge.03c",
                "pwge.04c",
                "pwge.05b",
                "pwg.06",
                "pwg.07",
                "pwg.08",
                "pwg.09a",
                "pwg.10"
            ]
            return dict(zip(filenames, chunks))
        func_map = {}
        func_map['maincpu'] = maincpu
        func_map['gfx'] = self._armwar_gfx
        func_map['audiocpu'] = self._armwar_audio
        func_map['qsound'] = self._armwar_qsound

        return helpers.build_rom(merged_contents, func_map)

    def _handle_pgear(self, merged_contents):

        def maincpu(contents):
            contents = contents[0x40:0x400040]
            chunks = transforms.equal_split(contents, num_chunks=8)
            filenames = [
                "pwgj.03a",
                "pwgj.04a",
                "pwgj.05a",
                "pwg.06",
                "pwg.07",
                "pwg.08",
                "pwg.09a",
                "pwg.10"
            ]
            return dict(zip(filenames, chunks))
        func_map = {}
        func_map['maincpu'] = maincpu
        func_map['gfx'] = self._armwar_gfx
        func_map['audiocpu'] = self._armwar_audio
        func_map['qsound'] = self._armwar_qsound

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Battle Circuit                                                               #
    ################################################################################

    def _batcir_gfx(self, contents):
        contents = contents[0x0800040:0x1800040]

        contents = capcom.common_gfx_deshuffle(contents)

        chunks = transforms.equal_split(contents, num_chunks=16)

        # Interleave each pair of chunks
        new_chunks = []
        for oddchunk, evenchunk in zip(chunks[0::2], chunks[1::2]):
            new_chunks.append(transforms.interleave(
                [oddchunk, evenchunk], word_size=8))
        chunks = new_chunks

        contents = transforms.merge(chunks)

        # Deinterleave the chunks into our 4 files
        chunks = transforms.deinterleave(contents, num_ways=4, word_size=2)
        filenames = [
            'btc.13m',
            'btc.15m',
            'btc.17m',
            'btc.19m'
        ]
        return dict(zip(filenames, chunks))

    def _batcir_audio(self, contents):
        chunks = []
        chunks.append(contents[0x1800040:0x1808040] +
                      contents[0x1810040:0x1828040])
        chunks.append(contents[0x1828040:0x1848040])
        filenames = [
            'btc.01',
            'btc.02'
        ]
        return dict(zip(filenames, chunks))

    def _batcir_qsound(self, contents):
        contents = contents[0x1850040:0x1C50040]
        chunks = transforms.equal_split(contents, num_chunks=2)
        chunks = transforms.swap_endian_all(chunks)
        filenames = [
            'btc.11m',
            'btc.12m'
        ]
        return dict(zip(filenames, chunks))

    def _handle_batcir(self, merged_contents):

        def maincpu(contents):
            contents = contents[0x40:0x380040]
            chunks = transforms.equal_split(contents, num_chunks=7)
            filenames = [
                "btce.03",
                "btce.04",
                "btce.05",
                "btce.06",
                "btc.07",
                "btc.08",
                "btc.09"
            ]
            return dict(zip(filenames, chunks))
        func_map = {}
        func_map['maincpu'] = maincpu
        func_map['gfx'] = self._batcir_gfx
        func_map['audiocpu'] = self._batcir_audio
        func_map['qsound'] = self._batcir_qsound
        
        return helpers.build_rom(merged_contents, func_map)

    def _handle_batcirj(self, merged_contents):

        def maincpu(contents):
            contents = contents[0x40:0x380040]
            chunks = transforms.equal_split(contents, num_chunks=7)
            filenames = [
                "btcj.03",
                "btcj.04",
                "btcj.05",
                "btcj.06",
                "btc.07",
                "btc.08",
                "btc.09"
            ]
            return dict(zip(filenames, chunks))
        func_map = {}
        func_map['maincpu'] = maincpu
        func_map['gfx'] = self._batcir_gfx
        func_map['audiocpu'] = self._batcir_audio
        func_map['qsound'] = self._batcir_qsound
        
        return helpers.build_rom(merged_contents, func_map)



'''Implementation of sf30ac: Street Fighter 30th Anniversary Collection'''
import glob
import logging
import os
from gex.lib.utils.blob import transforms
from gex.lib.utils.vendor import capcom
from gex.lib.contrib.bputil import BPListReader
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers

# TODO: Split class into a few smaller classes
# TODO: Add extraction of incomplete roms

logger = logging.getLogger('gextoolbox')

class SF30ACTask(BaseTask):
    '''Implements sf30ac: Street Fighter 30th Anniversary Collection'''
    _task_name = "sf30ac"
    _title = "Street Fighter 30th Anniversary Collection"
    _details_markdown = '''
This is reverse-engineered based on:
- The Japanese shell scripts in https://web.archive.org/web/20220213232038/http://blog.livedoor.jp/scrap_a/archives/22823395.html
- Valad Amoleo's https://github.com/ValadAmoleo/sf30ac-extractor/

Beyond the usual QSound dl-1425.bin and decryption keys, some of the CRC matches appear to be modified VROMs. The extraction is correct - 90%+ of the ROM matches - but details appear to be changed.

This script will extract and prep the ROMs. Some per-rom errata are in the notes below.

This will only extract the Japanese ROMs if you have SF30AC International Edition.
'''

    _default_input_folder = helpers.gen_steam_app_default_folder("Street Fighter 30th Anniversary Collection")
    _input_folder_desc = "SF30AC Steam folder"

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        for file_ref, file_metadata in self._metadata['in']['files'].items():
            resolved_file = self.read_datafile(in_dir, file_metadata)
            reader = BPListReader(resolved_file['contents'])
            is_international = resolved_file['version'] == "International Steam"
            parsed = reader.parse()

            handler_func = self.find_handler_func(file_ref)
            if parsed is not None and handler_func is not None:
                output_files = handler_func(parsed, is_international)

                for out_file_entry in output_files:
                    filename = out_file_entry['filename']
                    _ = self.verify_out_file(filename, out_file_entry['contents'])
                    out_path = os.path.join(out_dir, filename)
                    with open(out_path, "wb") as out_file:
                        out_file.write(out_file_entry['contents'])
            elif parsed is None:
                logger.warning("Could not find merged rom data in mbundle.")
            elif handler_func is None:
                logger.warning("Could not find matching handler function.")
        logger.info("Processing complete.")

    def _process_simm_common(self, simm_id, simm_prefix, simm_size_bytes):
        def process_simm(in_files):
            contents = in_files[simm_id]
            num_chunks = len(contents)//simm_size_bytes
            filenames = list(map(lambda x:f'{simm_prefix}-{simm_id}.{x}', range(0,num_chunks)))
            chunks = transforms.equal_split(contents, chunk_size = simm_size_bytes)
            return dict(zip(filenames, chunks))
        return process_simm

    def _deshuffle_gfx_common(self, filenames, num_interim_split, final_split = None):
        def gfx(in_files):
            contents = in_files['vrom']

            # This is weird... it's a bit shuffle, not byte-level and not a normal interleave
            contents = capcom.common_gfx_deshuffle(contents)

            # Split into even chunks
            chunks = transforms.equal_split(contents, num_chunks=num_interim_split)

            # Interleave each pair of chunks
            new_chunks = []
            for oddchunk,evenchunk in zip(chunks[0::2], chunks[1::2]):
                new_chunks.append(transforms.interleave([oddchunk, evenchunk], word_size=8))
            chunks = new_chunks

            # Merge the chunks back together
            contents = transforms.merge(chunks)

            # Deinterleave the chunks into our 4 files
            chunks = transforms.deinterleave(contents, num_ways = 4, word_size=2)

            # Do final split if provided
            if final_split:
                new_chunks = []
                for oldchunk in chunks:
                    new_chunks.extend(transforms.custom_split(oldchunk, final_split))
                chunks = new_chunks

            return dict(zip(filenames, chunks))
        return gfx

    def _cps2_gfx_deinterleave(self, contents, num_ways=4, word_size=2):
        interleave_group_length = num_ways * word_size
        num_interleave_groups = len(contents)//interleave_group_length
        contents = capcom.common_gfx_deshuffle(contents)
        temp_chunks = [bytearray() for i in range(num_ways)]
        for i in range(0, num_interleave_groups):
            offset = i * interleave_group_length
            interleave_group = contents[offset:offset+interleave_group_length]
            interleave_offset = 0
            for j in range(0, num_ways):
                interleave_end = interleave_offset + word_size
                temp_chunks[j] += interleave_group[interleave_offset:interleave_end]
                interleave_offset = interleave_end
        return temp_chunks

    ################################################################################
    # Street Fighter                                                               #
    ################################################################################

    def _handle_sf(self, mbundle_entries, is_international):
        func_map = {}
        out_files = []
        in_files = {}
        in_files['z80'] = mbundle_entries.get("StreetFighter.z80")
        in_files['alpha'] = mbundle_entries.get("StreetFighter.alpha.rom")
        in_files['68k'] = mbundle_entries.get("StreetFighter.u.68k")
        in_files['sprites'] = mbundle_entries.get("StreetFighter.sprites.rom")
        in_files['samples'] = mbundle_entries.get("StreetFighter.u.samples.rom")
        in_files['maps'] = mbundle_entries.get("StreetFighter.maps.rom")
        in_files['bplanes'] = mbundle_entries.get("StreetFighter.bplanes.rom")
        in_files['mplanes'] = mbundle_entries.get("StreetFighter.mplanes.rom")

        in_files['j-samples'] = mbundle_entries.get("StreetFighter.j.samples.rom")
        in_files['j-68k'] = mbundle_entries.get("StreetFighter.j.68k")

        bplanes_filenames = [
            "sf-39.2k",
            "sf-38.1k",
            "sf-41.4k",
            "sf-40.3k"
        ]
        func_map['bplanes'] = helpers.equal_split_helper('bplanes', bplanes_filenames)

        mplanes_filenames = [
            "sf-25.1d",
            "sf-28.1e",
            "sf-30.1g",
            "sf-34.1h",
            "sf-26.2d",
            "sf-29.2e",
            "sf-31.2g",
            "sf-35.2h"
        ]
        func_map['mplanes'] = helpers.equal_split_helper('mplanes', mplanes_filenames)

        sprites_filenames = [
            "sf-15.1m",
            "sf-16.2m",
            "sf-11.1k",
            "sf-12.2k",
            "sf-07.1h",
            "sf-08.2h",
            "sf-03.1f",
            "sf-17.3m",
            "sf-18.4m",
            "sf-13.3k",
            "sf-14.4k",
            "sf-09.3h",
            "sf-10.4h",
            "sf-05.3f"
        ]
        func_map['sprites'] = helpers.equal_split_helper('sprites', sprites_filenames)

        func_map['alpha'] = helpers.name_file_helper("alpha", "sf-27.4d")

        maps_filenames = [
            "sf-37.4h",
            "sf-36.3h",
            "sf-32.3g",
            "sf-33.4g"
        ]
        func_map['maps'] = helpers.equal_split_helper('maps', maps_filenames)

        func_map['z80'] = helpers.name_file_helper("z80", "sf-02.7k")

        logger.info("Processing SF common files...")
        common_file_map = helpers.process_rom_files(in_files, func_map)

        func_map = {}
        samples_filenames = [
            "sfu-00.1h",
            "sf-01.1k"
        ]
        func_map['samples'] = helpers.equal_split_helper('samples', samples_filenames)
        maincpu_filenames = [
            "sfd-19.2a",
            "sfd-22.2c",
            "sfd-20.3a",
            "sfd-23.3c",
            "sfd-21.4a",
            "sfd-24.4c"
        ]
        def sf_maincpu(in_file_name, filenames):
            def maincpu(in_files):
                contents = in_files[in_file_name]
                chunks = transforms.equal_split(contents, num_chunks = 3)

                new_chunks = []
                for oldchunk in chunks:
                    new_chunks.extend(transforms.deinterleave(oldchunk, num_ways=2, word_size=1))
                chunks = new_chunks

                return dict(zip(filenames, chunks))
            return maincpu
        func_map['maincpu'] = sf_maincpu('68k', maincpu_filenames)
        ph_files = {
            'mb7114h.12k': 0x100,
            'mb7114h.11h': 0x100,
            'mb7114h.12j': 0x100,
            'mmi-7603.13h': 0x020,
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        mame_name = "sf.zip"
        logger.info(f"Building {mame_name}...")
        out_files.append(
            {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
        )
        logger.info(f"Extracted {mame_name}.")

        # Handle international ROMs
        if is_international:
            logger.info("Japanese ROMs found, extracting...")
            func_map['maincpu'] = sf_maincpu('j-68k', maincpu_filenames)
            samples_filenames_j = [
                "sf-00.1h",
                "sf-01.1k"
            ]
            func_map['samples'] = helpers.equal_split_helper('j-samples', samples_filenames_j)
            ph_files_j = {
                'mb7114h.12j': 0x100,
                'sfb00.bin': 0x100,
                'sfb05.bin': 0x100,
                'mmi-7603.13h': 0x020,
                'sf_s.id8751h-8.14f': 0x1000
            }
            func_map['placeholders'] = helpers.placeholder_helper(ph_files_j)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name = "sfj.zip"
            logger.info(f"Building {mame_name}...")
            out_files.append(
                {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name}.")
        else:
            logger.info("Japanese ROMs not found, skipping.")

        return out_files

    ################################################################################
    #  Street Fighter 2                                                            #
    ################################################################################

    def _handle_sf2(self, mbundle_entries, is_international):
        func_map = {}
        out_files = []
        in_files = {}
        in_files['vrom'] = mbundle_entries.get('StreetFighterII.vrom')
        in_files['z80'] = mbundle_entries.get('StreetFighterII.z80')
        in_files['oki'] = mbundle_entries.get('StreetFighterII.oki')
        in_files['ub68k'] = mbundle_entries.get('StreetFighterII.ub.68k')
        in_files['ja-68k'] = mbundle_entries.get('StreetFighterII.ja.68k')
        in_files['jl-68k'] = mbundle_entries.get('StreetFighterII.jl.68k')

        # audiocpu
        audiocpu_filenames = [
            "sf2_9.12a"
        ]
        def audiocpu(in_files):
            contents = in_files['z80']
            return dict(zip(audiocpu_filenames, [contents]))
        func_map['audiocpu'] = audiocpu

        # gfx
        gfx_filenames = [
            "sf2_06.bin",
            "sf2_08.bin",
            "sf2_05.bin",
            "sf2_07.bin",
            "sf2_15.bin",
            "sf2_17.bin",
            "sf2_14.bin",
            "sf2_16.bin",
            "sf2_25.bin",
            "sf2_27.bin",
            "sf2_24.bin",
            "sf2_26.bin"
        ]
        def gfx(in_files):
            contents = in_files['vrom']
            chunks = transforms.equal_split(contents, num_chunks=3)

            new_chunks = []
            for oldchunk in chunks:
                new_chunks.extend(self._cps2_gfx_deinterleave(oldchunk, num_ways=4, word_size=2))
            chunks = new_chunks
            return dict(zip(gfx_filenames, chunks))
        func_map['gfx'] = gfx

        # oki
        oki_filenames = [
            'sf2_18.11c',
            'sf2_19.12c'
        ]
        def oki(in_files):
            chunks = transforms.equal_split(in_files['oki'], num_chunks=2)
            return dict(zip(oki_filenames, chunks))
        func_map['oki'] = oki

        ph_files = {
            'buf1': 0x117,
            'c632.ic1': 0x117,
            'ioa1': 0x117,
            'iob1.11d': 0x117,
            'prg1': 0x117,
            'rom1': 0x117,
            'sou1': 0x117,
            'stf29.1a': 0x117
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        logger.info("Processing SF2 common files...")
        common_file_map = helpers.process_rom_files(in_files, func_map)

        func_map = {}
        # maincpu
        maincpu_filenames = [
            'sf2_30a.bin',
            'sf2u.37b',
            'sf2_31a.bin',
            'sf2_38a.bin',
            'sf2_28a.bin',
            'sf2_35a.bin',
            'sf2_29a.bin',
            'sf2_36a.bin'
        ]
        def sf2_maincpu(in_file_name, filenames):
            def maincpu(in_files):
                contents = in_files[in_file_name]
                chunks = transforms.equal_split(contents, num_chunks = 4)

                new_chunks = []
                for oldchunk in chunks:
                    new_chunks.extend(transforms.deinterleave(oldchunk, num_ways=2, word_size=1))
                chunks = new_chunks

                return dict(zip(filenames, chunks))
            return maincpu
        func_map['maincpu'] = sf2_maincpu('ub68k', maincpu_filenames)
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        mame_name = "sf2ub.zip"
        logger.info(f"Building {mame_name}...")
        out_files.append(
            {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
        )
        logger.info(f"Extracted {mame_name}.")


        # Handle international ROMs
        if is_international:
            logger.info("Japanese ROMs found, extracting...")
            func_map = {}
            maincpu_filenames_ja = [
                'sf2j_30a.11e',
                'sf2u.37b',
                'sf2_31a.bin',
                'sf2_38a.bin',
                'sf2_28a.bin',
                'sf2_35a.bin',
                'sf2_29a.bin',
                'sf2_36a.bin'
            ]
            func_map['maincpu'] = sf2_maincpu('ja-68k', maincpu_filenames_ja)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name_ja = 'sf2ja.zip'
            logger.info(f"Building {mame_name_ja}...")
            out_files.append(
                {'filename': mame_name_ja, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name_ja}.")

            logger.info("Skipping sf2jl.zip as it needs a key but isn't in an old keyless MAME.")
        else:
            logger.info("Japanese ROMs not found, skipping.")


        return out_files

    ################################################################################
    # Street Fighter Alpha                                                         #
    ################################################################################

    def _handle_sfa(self, mbundle_entries, is_international):
        out_files = []
        func_map = {}
        in_files = {}
        in_files['vrom'] = mbundle_entries.get('StreetFighterAlpha.vrom')
        in_files['z80'] = mbundle_entries.get('StreetFighterAlpha.z80')
        in_files['qs'] = mbundle_entries.get('StreetFighterAlpha.qs')
        in_files['nv'] = mbundle_entries.get('StreetFighterAlpha.nv')
        in_files['u-68k'] = mbundle_entries.get('StreetFighterAlpha.u.68k')
        in_files['j-68k'] = mbundle_entries.get('StreetFighterAlpha.j.68k')
        in_files['jr2-68k'] = mbundle_entries.get('StreetFighterAlpha.jr2.68k')

        #vrom
        vrom_filenames = [
            "sfz.14m",
            "sfz.16m",
            "sfz.18m",
            "sfz.20m",
        ]
        func_map['vrom'] = self._deshuffle_gfx_common(vrom_filenames, 8)


        # z80
        z80_filenames = [
            'sfz.01',
            'sfz.02'
        ]
        def z80(in_files):
            chunks = transforms.equal_split(in_files['z80'], num_chunks=2)
            return dict(zip(z80_filenames, chunks))
        func_map['z80'] = z80

        # qsound
        qsound_filenames = [
            'sfz.11m',
            'sfz.12m'
        ]
        def qsound(in_files):
            chunks = transforms.equal_split(in_files['qs'], num_chunks=2)
            chunks = transforms.swap_endian_all(chunks)
            return dict(zip(qsound_filenames, chunks))
        func_map['qsound'] = qsound

        logger.info("Processing SFA1 common files...")
        common_file_map = helpers.process_rom_files(in_files, func_map)


        # maincpu
        maincpu_filenames = [
            'sfzu.03a',
            'sfz.04a',
            'sfz.05a',
            'sfz.06'
        ]
        def sfa_maincpu(in_file_name, filenames):
            def maincpu(in_files):
                contents = in_files[in_file_name]
                contents = transforms.swap_endian(contents)
                chunks = transforms.equal_split(contents, num_chunks = 4)

                return dict(zip(filenames, chunks))
            return maincpu
        func_map['maincpu'] = sfa_maincpu('u-68k', maincpu_filenames)
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        mame_name = "sfau.zip"
        logger.info(f"Building {mame_name}...")
        out_files.append(
            {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
        )
        logger.info(f"Extracted {mame_name}.")

        # Handle international ROMs
        if is_international:
            logger.info("Japanese ROMs found, extracting...")
            func_map = {}
            maincpu_filenames_j = [
                'sfzj.03c',
                'sfz.04b',
                'sfz.05a',
                'sfz.06'
            ]
            func_map['maincpu'] = sfa_maincpu('j-68k', maincpu_filenames_j)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name_j = 'sfzj.zip'
            logger.info(f"Building {mame_name_j}...")
            out_files.append(
                {'filename': mame_name_j, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name_j}.")

            func_map = {}
            maincpu_filenames_jr2 = [
                'sfzj.03b',
                'sfz.04a',
                'sfz.05a',
                'sfz.06'
            ]
            func_map['maincpu'] = sfa_maincpu('jr2-68k', maincpu_filenames_jr2)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name_jr2 = 'sfzjr2.zip'
            logger.info(f"Building {mame_name_jr2}...")
            out_files.append(
                {'filename': mame_name_jr2, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name_jr2}.")

        else:
            logger.info("Japanese ROMs not found, skipping.")

        return out_files

    ################################################################################
    # Street Fighter Alpha 2                                                       #
    ################################################################################

    def _handle_sfa2(self, mbundle_entries, is_international):
        out_files = []
        func_map = {}
        in_files = {}
        in_files['vrom'] = mbundle_entries.get('StreetFighterAlpha2.vrom')
        in_files['z80'] = mbundle_entries.get('StreetFighterAlpha2.z80')
        in_files['qs'] = mbundle_entries.get('StreetFighterAlpha2.qs')
        in_files['u168k'] = mbundle_entries.get('StreetFighterAlpha2.u1.68k')
        in_files['j-68k'] = mbundle_entries.get('StreetFighterAlpha2.j.68k')
        in_files['jr1-68k'] = mbundle_entries.get('StreetFighterAlpha2.jr1.68k')

        vrom_filenames = [
            "sz2.13m",
            "sz2.14m",
            "sz2.15m",
            "sz2.16m",
            "sz2.17m",
            "sz2.18m",
            "sz2.19m",
            "sz2.20m"
        ]
        func_map['vrom'] = self._deshuffle_gfx_common(
            vrom_filenames,
            20,
            final_split = [0x400000, 0x100000]
        )

        # z80
        z80_filenames = [
            'sz2.01a',
            'sz2.02a'
        ]
        def z80(in_files):
            chunks = transforms.equal_split(in_files['z80'], num_chunks=2)
            return dict(zip(z80_filenames, chunks))
        func_map['z80'] = z80

        # qsound
        qsound_filenames = [
            'sz2.11m',
            'sz2.12m'
        ]
        def qsound(in_files):
            chunks = transforms.equal_split(in_files['qs'], num_chunks=2)
            chunks = transforms.swap_endian_all(chunks)
            return dict(zip(qsound_filenames, chunks))
        func_map['qsound'] = qsound

        logger.info("Processing SFA2 common files...")
        common_file_map = helpers.process_rom_files(in_files, func_map)

        func_map = {}
        maincpu_filenames = [
            "sz2u.03",
            "sz2u.04",
            "sz2u.05",
            "sz2u.06",
            "sz2u.07",
            "sz2u.08"
        ]
        def sfa2_maincpu(in_file_name, filenames):
            def maincpu(in_files):
                contents = in_files[in_file_name]
                contents = transforms.swap_endian(contents)
                chunks = transforms.equal_split(contents, num_chunks = 6)

                return dict(zip(filenames, chunks))
            return maincpu
        func_map['maincpu'] = sfa2_maincpu('u168k', maincpu_filenames)
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        mame_name = "sfa2u.zip"
        logger.info(f"Building {mame_name}...")
        out_files.append(
            {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
        )
        logger.info(f"Extracted {mame_name}.")

        # Handle international ROMs
        if is_international:
            logger.info("Japanese ROMs found, extracting...")
            func_map = {}
            maincpu_filenames_j = [
                "sz2j.03a",
                "sz2j.04a",
                "sz2.05a",
                "sz2.06",
                "sz2j.07a",
                "sz2.08"
            ]
            func_map['maincpu'] = sfa2_maincpu('j-68k', maincpu_filenames_j)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name_j = 'sfz2j.zip'
            logger.info(f"Building {mame_name_j}...")
            out_files.append(
                {'filename': mame_name_j, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name_j}.")

            logger.info("Skipping sfz2jr1.zip as it needs a key but isn't in an old keyless MAME.")

        else:
            logger.info("Japanese ROMs not found, skipping.")


        return out_files

    ################################################################################
    # Street Fighter Alpha 3                                                       #
    ################################################################################

    def _handle_sfa3(self, mbundle_entries, is_international):
        out_files = []
        func_map = {}
        in_files = {}
        in_files['vrom'] = mbundle_entries.get('StreetFighterAlpha3.vrom')
        in_files['z80'] = mbundle_entries.get('StreetFighterAlpha3.z80')
        in_files['qs'] = mbundle_entries.get('StreetFighterAlpha3.qs')

        in_files['u68k'] = mbundle_entries.get('StreetFighterAlpha3.u.68k')

        in_files['j-68k'] = mbundle_entries.get('StreetFighterAlpha3.j.68k')
        in_files['jr2-68k'] = mbundle_entries.get('StreetFighterAlpha3.jr2.68k')

        vrom_filenames = [
            "sz3.13m",
            "sz3.14m",
            "sz3.15m",
            "sz3.16m",
            "sz3.17m",
            "sz3.18m",
            "sz3.19m",
            "sz3.20m"
        ]
        func_map['vrom'] = self._deshuffle_gfx_common(
            vrom_filenames,
            32,
            final_split = [0x400000, 0x400000]
        )

        # z80
        z80_filenames = [
            'sz3.01',
            'sz3.02'
        ]
        def z80(in_files):
            chunks = transforms.equal_split(in_files['z80'], num_chunks=2)
            return dict(zip(z80_filenames, chunks))
        func_map['z80'] = z80

        # qsound
        qsound_filenames = [
            'sz3.11m',
            'sz3.12m'
        ]
        def qsound(in_files):
            chunks = transforms.equal_split(in_files['qs'], num_chunks=2)
            chunks = transforms.swap_endian_all(chunks)
            return dict(zip(qsound_filenames, chunks))
        func_map['qsound'] = qsound

        logger.info("Processing SFA3 common files...")
        common_file_map = helpers.process_rom_files(in_files, func_map)


        func_map = {}
        maincpu_filenames = [
            "sz3u.03c",
            "sz3u.04c",
            "sz3.05c",
            "sz3.06c",
            "sz3.07c",
            "sz3.08c",
            "sz3.09c",
            "sz3.10b"
        ]
        def sfa3_maincpu(in_file_name, filenames):
            def maincpu(in_files):
                contents = in_files[in_file_name]
                contents = transforms.swap_endian(contents)
                chunks = transforms.equal_split(contents, num_chunks = 8)

                return dict(zip(filenames, chunks))
            return maincpu
        func_map['maincpu'] = sfa3_maincpu('u68k', maincpu_filenames)
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        mame_name = "sfa3u.zip"
        logger.info(f"Building {mame_name}...")
        out_files.append(
            {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
        )
        logger.info(f"Extracted {mame_name}.")

        # Handle international ROMs
        if is_international:
            logger.info("Japanese ROMs found, extracting...")
            func_map = {}
            maincpu_filenames_j = [
                "sz3j.03c",
                "sz3j.04c",
                "sz3.05c",
                "sz3.06c",
                "sz3.07c",
                "sz3.08c",
                "sz3.09c",
                "sz3.10b"
            ]
            func_map['maincpu'] = sfa3_maincpu('j-68k', maincpu_filenames_j)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name_j = 'sfz3j.zip'
            logger.info(f"Building {mame_name_j}...")
            out_files.append(
                {'filename': mame_name_j, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name_j}.")

            func_map = {}
            maincpu_filenames_jr2 = [
                "sz3j.03",
                "sz3j.04",
                "sz3.05",
                "sz3.06",
                "sz3.07",
                "sz3.08",
                "sz3.09",
                "sz3.10"
            ]
            func_map['maincpu'] = sfa3_maincpu('jr2-68k', maincpu_filenames_jr2)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name_jr2 = 'sfz3jr2.zip'
            logger.info(f"Building {mame_name_jr2}...")
            out_files.append(
                {'filename': mame_name_jr2, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name_jr2}.")

        else:
            logger.info("Japanese ROMs not found, skipping.")

        return out_files

    ################################################################################
    # Street Fighter 3                                                             #
    ################################################################################

    def _sf3_common(self, mbundle_entries, in_bios_filename, in_simm_bank_files, simm_prefix, bios_filename, mame_name):
        out_files = []
        func_map = {}
        in_files = {}
        simm_size = 2*1024*1024
        for simm_bank_num, simm_filename in in_simm_bank_files.items():
            bank_name = f'simm{simm_bank_num}'
            in_files[bank_name] = mbundle_entries.get(simm_filename)
            func_map[bank_name] = self._process_simm_common(bank_name, simm_prefix, simm_size)

        in_files['bios'] = mbundle_entries.get(in_bios_filename)
        func_map['bios'] = helpers.name_file_helper("bios", bios_filename)

        out_files.append({'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)})

        return out_files

    def _handle_sf3(self, mbundle_entries, _):
        in_prefix = "StreetFighterIII"
        in_simm_bank_nums = [1, 3, 4, 5]
        in_simm_files = dict(zip(in_simm_bank_nums,
            list(map(lambda x:f'{in_prefix}.s{x}', in_simm_bank_nums))))
        in_bios_file = f'{in_prefix}.bios'
        return self._sf3_common(
            mbundle_entries,
            in_bios_file,
            in_simm_files,
            'sfiii',
            "sfiii_asia_nocd.29f400.u2",
            'sfiiina.zip'
        )

    ################################################################################
    # Street Fighter 3 2nd Impact                                                  #
    ################################################################################

    def _handle_sf3_2i(self, mbundle_entries, _):
        in_prefix = "StreetFighterIII_2ndImpact"
        in_simm_bank_nums = list(range(1,6))
        in_simm_files = dict(zip(in_simm_bank_nums,
            list(map(lambda x:f'{in_prefix}.s{x}', in_simm_bank_nums))))
        in_bios_file = f'{in_prefix}.bios'
        return self._sf3_common(
            mbundle_entries,
            in_bios_file,
            in_simm_files,
            'sfiii2',
            "sfiii2_asia_nocd.29f400.u2",
            'sfiii2n.zip'
        )

    ################################################################################
    # Street Fighter 3 3rd Strike                                                  #
    ################################################################################

    def _handle_sf3_3s(self, mbundle_entries, _):
        in_prefix = "StreetFighterIII_3rdStrike"
        in_simm_files = {
            1: f'{in_prefix}.r1.s1',
            2: f'{in_prefix}.r1.s2',
            3: f'{in_prefix}.s3',
            4: f'{in_prefix}.s4',
            5: f'{in_prefix}.s5',
            6: f'{in_prefix}.s6'
        }
        in_bios_file = f'{in_prefix}.bios'
        return self._sf3_common(
            mbundle_entries,
            in_bios_file,
            in_simm_files,
            'sfiii3',
            "sfiii3_japan_nocd.29f400.u2",
            'sfiii3nr1.zip'
        )

    ################################################################################
    # Street Fighter 2 Championship Edition                                        #
    ################################################################################

    def _handle_sf2ce(self, mbundle_entries, is_international):
        func_map = {}
        out_files = []
        in_files = {}
        in_files['vrom'] = mbundle_entries.get('StreetFighterII_CE.vrom')
        in_files['z80'] = mbundle_entries.get('StreetFighterII_CE.z80')
        in_files['oki'] = mbundle_entries.get('StreetFighterII_CE.oki')
        in_files['68k'] = mbundle_entries.get('StreetFighterII_CE.ua.68k')
        in_files['jb-68k'] = mbundle_entries.get('StreetFighterII_CE.jb.68k')

        # audiocpu
        audiocpu_filenames = [
            "s92_09.bin"
        ]
        def audiocpu(in_files):
            contents = in_files['z80']
            return dict(zip(audiocpu_filenames, [contents]))
        func_map['audiocpu'] = audiocpu

        # gfx
        gfx_filenames = [
            "s92_01.bin",
            "s92_02.bin",
            "s92_03.bin",
            "s92_04.bin",
            "s92_05.bin",
            "s92_06.bin",
            "s92_07.bin",
            "s92_08.bin",
            "s92_10.bin",
            "s92_11.bin",
            "s92_12.bin",
            "s92_13.bin"
        ]
        def gfx(in_files):
            contents = in_files['vrom']
            chunks = transforms.equal_split(contents, num_chunks=3)

            new_chunks = []
            for oldchunk in chunks:
                new_chunks.extend(self._cps2_gfx_deinterleave(oldchunk, num_ways=4, word_size=2))
            chunks = new_chunks
            return dict(zip(gfx_filenames, chunks))
        func_map['gfx'] = gfx

        # oki
        oki_filenames = [
            's92_18.bin',
            's92_19.bin'
        ]
        def oki(in_files):
            chunks = transforms.equal_split(in_files['oki'], num_chunks=2)
            return dict(zip(oki_filenames, chunks))
        func_map['oki'] = oki

        ph_files = {
            'bprg1.11d': 0x117,
            'buf1': 0x117,
            'c632.ic1': 0x117,
            'ioa1': 0x117,
            'iob1.12d': 0x117,
            'prg1': 0x117,
            'rom1': 0x117,
            'sou1': 0x117,
            'ioc1.ic7': 0x104
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        logger.info("Processing SF2CE common files...")
        common_file_map = helpers.process_rom_files(in_files, func_map)


        func_map = {}
        # maincpu
        maincpu_filenames = [
            "s92u-23a",
            "sf2ce.22",
            "s92_21a.bin"
        ]
        def sf2ce_maincpu(in_file_name, filenames):
            def maincpu(in_files):
                contents = in_files[in_file_name]
                chunks = transforms.equal_split(contents, num_chunks = 3)
                chunks = transforms.swap_endian_all(chunks)
                return dict(zip(filenames, chunks))
            return maincpu
        func_map['maincpu'] = sf2ce_maincpu('68k', maincpu_filenames)
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        mame_name = "sf2ceua.zip"
        logger.info(f"Building {mame_name}...")
        out_files.append(
            {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
        )
        logger.info(f"Extracted {mame_name}.")

        # Handle international ROMs
        if is_international:
            logger.info("Japanese ROMs found, extracting...")
            func_map = {}
            maincpu_filenames = [
                "s92j_23b.bin",
                "s92j_22b.bin",
                "s92_21a.bin"
            ]
            func_map['maincpu'] = sf2ce_maincpu('jb-68k', maincpu_filenames)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name = 'sf2cej.zip'
            logger.info(f"Building {mame_name}...")
            out_files.append(
                {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name}.")

            logger.info("Skipping sf2cej.zip as it needs a key but isn't in an old keyless MAME.")
        else:
            logger.info("Japanese ROMs not found, skipping.")

        return out_files

    ################################################################################
    # Street Fighter 2 Hyper Fighting                                              #
    ################################################################################

    def _handle_sf2hf(self, mbundle_entries, is_international):
        func_map = {}
        out_files = []
        in_files = {}
        in_files['vrom'] = mbundle_entries.get('StreetFighterII_HF.u.vrom')
        in_files['z80'] = mbundle_entries.get('StreetFighterII_HF.z80')
        in_files['oki'] = mbundle_entries.get('StreetFighterII_HF.oki')
        in_files['68k'] = mbundle_entries.get('StreetFighterII_HF.u.68k')
        in_files['j-vrom'] = mbundle_entries.get('StreetFighterII_HF.j.p16.p32.vrom')
        in_files['j-68k'] = mbundle_entries.get('StreetFighterII_HF.j.68k')

        # audiocpu
        audiocpu_filenames = [
            "s92_09.bin"
        ]
        def audiocpu(in_files):
            contents = in_files['z80']
            return dict(zip(audiocpu_filenames, [contents]))
        func_map['audiocpu'] = audiocpu

        # oki
        oki_filenames = [
            's92_18.bin',
            's92_19.bin'
        ]
        def oki(in_files):
            chunks = transforms.equal_split(in_files['oki'], num_chunks=2)
            return dict(zip(oki_filenames, chunks))
        func_map['oki'] = oki

        ph_files = {
            'bprg1.11d': 0x117,
            'buf1': 0x117,
            'c632.ic1': 0x117,
            'ioa1': 0x117,
            'iob1.12d': 0x117,
            'prg1': 0x117,
            'rom1': 0x117,
            'sou1': 0x117,
            'ioc1.ic7': 0x104
        }
        func_map['placeholders'] = helpers.placeholder_helper(ph_files)

        logger.info("Processing SF2HF common files...")
        common_file_map = helpers.process_rom_files(in_files, func_map)

        func_map = {}
        # maincpu
        maincpu_filenames = [
            "sf2_23a",
            "sf2_22.bin",
            "sf2_21.bin"
        ]
        def sf2hf_maincpu(in_file_name, filenames):
            def maincpu(in_files):
                contents = in_files[in_file_name]
                chunks = transforms.equal_split(contents, num_chunks = 3)
                chunks = transforms.swap_endian_all(chunks)
                return dict(zip(filenames, chunks))
            return maincpu
        func_map['maincpu'] = sf2hf_maincpu('68k', maincpu_filenames)

        # gfx
        gfx_filenames = [
            "s92_01.bin",
            "s92_02.bin",
            "s92_03.bin",
            "s92_04.bin",
            "s92_05.bin",
            "s92_06.bin",
            "s92_07.bin",
            "s92_08.bin",
            "s2t_10.bin",
            "s2t_11.bin",
            "s2t_12.bin",
            "s2t_13.bin"
        ]
        def sf2hf_gfx(in_file_name, filenames):
            def gfx(in_files):
                contents = in_files[in_file_name]
                chunks = transforms.equal_split(contents, num_chunks=3)

                new_chunks = []
                for oldchunk in chunks:
                    new_chunks.extend(
                        self._cps2_gfx_deinterleave(oldchunk, num_ways=4, word_size=2)
                    )
                chunks = new_chunks
                return dict(zip(filenames, chunks))
            return gfx
        func_map['gfx'] = sf2hf_gfx('vrom', gfx_filenames)
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        mame_name = "sf2t.zip"
        logger.info(f"Building {mame_name}...")
        out_files.append(
            {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
        )
        logger.info(f"Extracted {mame_name}.")

        # Handle international ROMs
        if is_international:
            logger.info("Japanese ROMs found, extracting...")
            func_map = {}
            maincpu_filenames = [
                "s2tj_23.bin",
                "sft_22.bin",
                "sft_21.bin"
            ]
            func_map['maincpu'] = sf2hf_maincpu('j-68k', maincpu_filenames)
            # gfx
            gfx_filenames = [
                "s92_01.bin",
                "s92_02.bin",
                "s92_03.bin",
                "s92_04.bin",
                "s92_05.bin",
                "s92_06.bin",
                "s92_07.bin",
                "s92_08.bin",
                "s2t_10.bin",
                "s2t_11.bin",
                "s2t_12.bin",
                "s2t_13.bin"
            ]
            func_map['gfx'] = sf2hf_gfx('j-vrom', gfx_filenames)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name_jb = 'sf2tj.zip'
            logger.info(f"Building {mame_name_jb}...")
            out_files.append(
                {'filename': mame_name_jb, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name_jb}.")
        else:
            logger.info("Japanese ROMs not found, skipping.")


        return out_files

    ################################################################################
    # Super Street Fighter 2                                                       #
    ################################################################################

    def _handle_ssf2(self, mbundle_entries, is_international):
        func_map = {}
        out_files = []
        in_files = {}
        in_files['vrom'] = mbundle_entries.get('SuperStreetFighterII.vrom')
        in_files['z80'] = mbundle_entries.get('SuperStreetFighterII.z80')
        in_files['qsound'] = mbundle_entries.get('SuperStreetFighterII.qs')
        in_files['68k'] = mbundle_entries.get('SuperStreetFighterII.u.68k')
        in_files['j-68k'] = mbundle_entries.get('SuperStreetFighterII.j.68k')
        in_files['jr1-68k'] = mbundle_entries.get('SuperStreetFighterII.jr1.68k')

        # audiocpu
        audiocpu_filenames = [
            "ssf.01"
        ]
        def audiocpu(in_files):
            contents = in_files['z80']
            return dict(zip(audiocpu_filenames, [contents]))
        func_map['audiocpu'] = audiocpu


        vrom_filenames = [
            "ssf.13m",
            "ssf.14m",
            "ssf.15m",
            "ssf.16m",
            "ssf.17m",
            "ssf.18m",
            "ssf.19m",
            "ssf.20m"
        ]
        func_map['vrom'] = self._deshuffle_gfx_common(
            vrom_filenames,
            12,
            final_split = [0x200000, 0x100000]
        )

        # qsound
        qsound_filenames = [
            "ssf.q01",
            "ssf.q02",
            "ssf.q03",
            "ssf.q04",
            "ssf.q05",
            "ssf.q06",
            "ssf.q07",
            "ssf.q08"
        ]
        def qsound(in_files):
            chunks = transforms.equal_split(in_files['qsound'], num_chunks=8)
            return dict(zip(qsound_filenames, chunks))
        func_map['qsound'] = qsound

        logger.info("Processing SSF2 common files...")
        common_file_map = helpers.process_rom_files(in_files, func_map)

        func_map = {}
        # maincpu
        maincpu_filenames = [
            "ssfu.03a",
            "ssfu.04a",
            "ssfu.05",
            "ssfu.06",
            "ssfu.07"
        ]
        def ssf2_maincpu(in_file_name, filenames):
            def maincpu(in_files):
                contents = in_files[in_file_name]
                chunks = transforms.equal_split(contents, num_chunks = 5)
                chunks = transforms.swap_endian_all(chunks)
                return dict(zip(filenames, chunks))
            return maincpu
        func_map['maincpu'] = ssf2_maincpu('68k', maincpu_filenames)
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        mame_name = "ssf2u.zip"
        logger.info(f"Building {mame_name}...")
        out_files.append(
            {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
        )
        logger.info(f"Extracted {mame_name}.")

        # Handle international ROMs
        if is_international:
            logger.info("Japanese ROMs found, extracting...")
            func_map = {}
            maincpu_filenames = [
                "ssfj.03b",
                "ssfj.04a",
                "ssfj.05",
                "ssfj.06b",
                "ssfj.07"
            ]
            func_map['maincpu'] = ssf2_maincpu('j-68k', maincpu_filenames)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name = 'ssf2j.zip'
            logger.info(f"Building {mame_name}...")
            out_files.append(
                {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name}.")
            func_map = {}
            maincpu_filenames = [
                "ssfj.03a",
                "ssfj.04a",
                "ssfj.05",
                "ssfj.06",
                "ssfj.07"
            ]
            func_map['maincpu'] = ssf2_maincpu('jr1-68k', maincpu_filenames)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name = 'ssf2jr1.zip'
            logger.info(f"Building {mame_name}...")
            out_files.append(
                {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name}.")
        else:
            logger.info("Japanese ROMs not found, skipping.")

        return out_files

    ################################################################################
    # Super Street Fighter 2 Turbo                                                 #
    ################################################################################

    def _handle_ssf2t(self, mbundle_entries, is_international):
        func_map = {}
        out_files = []
        in_files = {}
        in_files['vrom'] = mbundle_entries.get('SuperStreetFighterIITurbo.vrom')
        in_files['z80'] = mbundle_entries.get('SuperStreetFighterIITurbo.z80')
        in_files['qsound'] = mbundle_entries.get('SuperStreetFighterIITurbo.qs')
        in_files['68k'] = mbundle_entries.get('SuperStreetFighterIITurbo.u.68k')
        in_files['j-68k'] = mbundle_entries.get('SuperStreetFighterIITurbo.j.68k')
        in_files['jr1-68k'] = mbundle_entries.get('SuperStreetFighterIITurbo.jr1.68k')

        # audiocpu
        audiocpu_filenames = [
            "sfx.01",
            "sfx.02"
        ]
        def audiocpu(in_files):
            contents = in_files['z80']
            chunks = transforms.equal_split(contents, num_chunks = len(audiocpu_filenames))
            return dict(zip(audiocpu_filenames, chunks))
        func_map['audiocpu'] = audiocpu

        vrom_filenames = [
            "sfx.13m",
            "sfx.14m",
            "sfx.21m",
            "sfx.15m",
            "sfx.16m",
            "sfx.23m",
            "sfx.17m",
            "sfx.18m",
            "sfx.25m",
            "sfx.19m",
            "sfx.20m",
            "sfx.27m"
        ]
        func_map['vrom'] = self._deshuffle_gfx_common(
            vrom_filenames,
            16,
            final_split = [0x200000, 0x100000, 0x100000]
        )

        # qsound
        qsound_filenames = [
            "sfx.11m",
            "sfx.12m",
        ]
        def qsound(in_files):
            chunks = transforms.equal_split(in_files['qsound'], num_chunks=2)
            return dict(zip(qsound_filenames, chunks))
        func_map['qsound'] = qsound

        logger.info("Processing SSFT2 common files...")
        common_file_map = helpers.process_rom_files(in_files, func_map)

        func_map = {}
        # maincpu
        maincpu_filenames = [
            "sfxu.03e",
            "sfxu.04a",
            "sfxu.05",
            "sfxu.06b",
            "sfxu.07a",
            "sfxu.08",
            "sfx.09"
        ]
        def ssf2t_maincpu(in_file_name, filenames):
            def maincpu(in_files):
                contents = in_files[in_file_name]
                chunks = transforms.equal_split(contents, num_chunks = len(maincpu_filenames))
                chunks = transforms.swap_endian_all(chunks)
                return dict(zip(filenames, chunks))
            return maincpu
        func_map['maincpu'] = ssf2t_maincpu('68k', maincpu_filenames)
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        mame_name = "ssf2tu.zip"
        logger.info(f"Building {mame_name}...")
        out_files.append(
            {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
        )
        logger.info(f"Extracted {mame_name}.")

        # Handle international ROMs
        if is_international:
            logger.info("Japanese ROMs found, extracting...")
            func_map = {}
            maincpu_filenames = [
                "sfxj.03c",
                "sfxj.04a",
                "sfxj.05",
                "sfxj.06a",
                "sfxj.07",
                "sfxj.08",
                "sfx.09"
            ]
            func_map['maincpu'] = ssf2t_maincpu('j-68k', maincpu_filenames)
            func_map['common'] = helpers.existing_files_helper(common_file_map)
            mame_name = 'ssf2xj.zip'
            logger.info(f"Building {mame_name}...")
            out_files.append(
                {'filename': mame_name, 'contents': helpers.build_rom(in_files, func_map)}
            )
            logger.info(f"Extracted {mame_name}.")

            logger.info("Skipping ssf2xjr1.zip as it needs a key but isn't in an old keyless MAME.")
        else:
            logger.info("Japanese ROMs not found, skipping.")

        return out_files



'''Convenience wrappers for task implementation'''
import io
import os
import zipfile
from gex.lib.utils.blob import transforms
import gex.lib.archive.zip as ziphelper

STEAM_APP_ROOT = r"C:\Program Files (x86)\Steam\steamapps\common"
def gen_steam_app_default_folder(app_folder, library_root=STEAM_APP_ROOT):
    '''Convenience function to get a Steam App folder'''
    return os.path.join(library_root, app_folder)

def build_rom(in_files, func_map):
    '''Convenience function to run both process_rom_files and build_zip together'''
    file_map = process_rom_files(in_files, func_map)
    return build_zip(file_map)

def process_rom_files(in_files, func_map):
    '''Look at a list of content files and run a set of transform functions on them'''
    file_map = {}
    for func in func_map.values():
        file_map.update(func(in_files))
    return file_map

def create_combined_file_map(*file_maps):
    '''Create a combined file map from 2 or more existing maps'''
    new_map = {}
    for file_map in file_maps:
        new_map.update(file_map)
    return new_map

def build_zip(file_map):
    '''Build a zip file from a dictionary of paths to contents'''
    new_contents = io.BytesIO()
    with zipfile.ZipFile(new_contents, "w", compression=zipfile.ZIP_DEFLATED) as new_archive:
        for name, data in file_map.items():
            new_archive.writestr(name, data)
    return new_contents.getvalue()

def existing_files_helper(file_map):
    '''Func map helper to reuse a file map'''
    def existing_files(*_):
        return file_map
    return existing_files

def equal_split_helper(in_file_ref, filenames):
    '''Func map helper for transforms.equal_split'''
    def split(in_files):
        contents = in_files[in_file_ref]
        chunks = transforms.equal_split(contents, num_chunks = len(filenames))
        return dict(zip(filenames, chunks))
    return split

def custom_split_helper(in_file_ref, name_size_map):
    '''Func map helper for transforms.custom_split'''
    def split(in_files):
        contents = in_files[in_file_ref]
        chunks = transforms.custom_split(contents, list(name_size_map.values()))
        return dict(zip(name_size_map.keys(), chunks))
    return split

def deinterleave_helper(in_file_name, filenames, num_ways, word_size):
    '''Func map helper for deinterleaving a file'''
    def deinterleave(in_files):
        contents = in_files[in_file_name]
        chunks = transforms.deinterleave(contents, num_ways=num_ways, word_size=word_size)
        return dict(zip(filenames, chunks))
    return deinterleave

def name_file_helper(in_file_ref, filename):
    '''Func map helper for renaming a file'''
    def rename_from(in_files):
        return {filename: in_files[in_file_ref]}
    return rename_from

def splice_out_helper(start, length=None, end=None):
    '''Func map helper for transforms.splice_out'''
    def splice_func(contents):
        return transforms.splice_out(contents, start, length, end)
    return splice_func

def slice_helper(start=0, length=None, end=None):
    '''Func map helper for slicing a blob'''
    if length is None and end is None:
        raise Exception("Splice out needs a length or end value, but received neither.")
    elif length is not None and end is not None:
        raise Exception("Splice out needs a length or end value, but received both.")
    elif end is None:
        end = start + length
    def slice_func(contents):
        return contents[start:end]
    return slice_func

def placeholder_helper(file_map):
    '''Func map helper for making empty placeholder files'''
    def create_placeholders(_):
        out_files = {}
        for filename, size in file_map.items():
            out_files[filename] = bytes(size*b'\0')
        return out_files
    return create_placeholders

def common_picker_helper(common_file_map, src_name, dst_name=None):
    '''Func map helper for picking/renaming a single file from an existing common map'''
    def pick(_):
        out_files = {}

        content = common_file_map.get(src_name)
        filename = dst_name if not None else src_name

        out_files[filename] = content

        return out_files
    return pick

def common_rename_helper(common_file_map, rename_map):
    '''Func map helper for picking/renaming a single file from an existing common map'''
    def pick(_):
        out_files = {}

        for src_name, dst_name in rename_map.items():
            out_files[dst_name] = common_file_map.get(src_name)

        return out_files
    return pick

def pull_files_from_archive(archive_data, archive_type, file_map):
    out_files = []

    archive_contents = None
    if archive_type == "zip":
        archive_contents = ziphelper.extract(archive_data)

    for destination_name, archive_name in file_map.items():
        out_files.append({
            'filename': destination_name,
            'contents': archive_contents[archive_name]['contents']
        })

    return out_files


'''Extraction code for Atari Arcade ROMs'''
import logging
import os
from gex.lib.tasks import helpers
from gex.lib.utils.blob import transforms
from gex.lib.utils import gfx_rebuilder

logger = logging.getLogger('gextoolbox')

def _handle_mhavoc(in_files, game_desc):
    zip_files = {}

    # Major Havoc.bin
    contents = in_files.get('Major Havoc.bin')['contents']
    chunks = transforms.custom_split(contents, [0x1000, 0x4000, 0x4000])
    zip_files['136025.210'] = transforms.pad(chunks[0], 0x2000)
    zip_files['136025.216'] = chunks[1]
    zip_files['136025.217'] = chunks[2]

    # Major Havoc alpha banks.bin
    contents = in_files.get('Major Havoc alpha banks.bin')['contents']
    chunks = transforms.equal_split(contents, 2)
    zip_files['136025.215'] = chunks[0]
    zip_files['136025.318'] = chunks[1]

    # Major Havoc gamma.bin
    contents = in_files.get('Major Havoc gamma.bin')['contents']
    zip_files['136025.108'] = contents

    # Major Havoc vector banks.bin
    contents = in_files.get('Major Havoc vector banks.bin')['contents']
    chunks = transforms.equal_split(contents, 2)
    zip_files['136025.106'] = chunks[0]
    zip_files['136025.107'] = chunks[1]

    return [{
        'filename': game_desc['filename'],
        'contents': helpers.build_zip(zip_files)
    }]

def _handle_montecar(in_files, game_desc):
    zip_files = {}

    # MonteCarlo.bin
    contents = in_files.get('MonteCarlo.bin')['contents']
    chunks = transforms.equal_split(contents, 4)
    zip_files['35763-01.h1'] = chunks[0]
    zip_files['35763-01.f1'] = chunks[1]
    zip_files['35763-01.d1'] = chunks[2]
    zip_files['35763-01.c1'] = chunks[3]

    # MonteCarloColors.bin
    contents = in_files.get('MonteCarloColors.bin')['contents']
    zip_files['35785-01.e7'] = contents

    # Bitmaps
    bitmaps = [
        "MonteCarloSprites1.bmp",
        "MonteCarloSprites2.bmp",
        "MonteCarloTiles1.bmp",
        "MonteCarloTiles2.bmp"
    ]
    for bitmap_file in bitmaps:
        contents = in_files.get(bitmap_file)['contents']
        zip_files[bitmap_file] = contents

    return [{
        'filename': game_desc['filename'],
        'contents': helpers.build_zip(zip_files)
    }]

def _handle_redbaron(in_files, game_desc):
    zip_files = {}

    # Red Baron.bin
    contents = in_files.get('Red Baron.bin')['contents']
    chunks = transforms.equal_split(contents, 9)
    zip_files['35763-01.h1'] = chunks[0]

    zip_files["037587.01"] = transforms.merge([chunks[0] + chunks[2]])
    zip_files["037000.01e"] = chunks[1]
    zip_files["036998.01e"] = chunks[3]
    zip_files["036997.01e"] = chunks[4]
    zip_files["036996.01e"] = chunks[5]
    zip_files["036995.01e"] = chunks[6]
    zip_files["037006.01e"] = chunks[7]
    zip_files["037007.01e"] = chunks[8]

    return [{
        'filename': game_desc['filename'],
        'contents': helpers.build_zip(zip_files)
    }]

def _handle_sprint2(in_files, game_desc):
    zip_files = {}

    # Sprint2.bin
    contents = in_files.get('Sprint2.bin')['contents']
    chunks = transforms.equal_split(contents, 4)
    zip_files['6290-01.b1'] = chunks[0]
    zip_files['6291-01.c1'] = chunks[1]
    zip_files['6404.d1'] = chunks[2]
    zip_files['6405.e1'] = chunks[3]

    # Sprint2Sprites.bmp
    contents = in_files.get('Sprint2Sprites.bmp')['contents']
    contents = gfx_rebuilder.reverse_bmp(contents)
    sprint2_car_layout = {
        'width': 16,
        'height': 8,
        'total': 32,
        'planes': 1,
        'planeoffset': [0],
        'xoffset': [0x7, 0x6, 0x5, 0x4, 0x3, 0x2, 0x1, 0x0,
                0xf, 0xe, 0xd, 0xc, 0xb, 0xa, 0x9, 0x8],
        'yoffset': [0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0x60, 0x70],
        'charincrement': 0x80
    }
    contents = gfx_rebuilder.reencode_gfx(contents, sprint2_car_layout)
    chunks = transforms.deinterleave_nibble(contents, 2)
    zip_files['6399-01.j6'] = chunks[0]
    zip_files['6398-01.k6'] = chunks[1]

    # Sprint2Tiles.bmp
    contents = in_files.get('Sprint2Tiles.bmp')['contents']
    contents = gfx_rebuilder.reverse_bmp(contents)
    sprint2_tile_layout = {
        'width': 8,
        'height': 8,
        'total': 64,
        'planes': 1,
        'planeoffset': [0],
        'xoffset': [0, 1, 2, 3, 4, 5, 6, 7],
        'yoffset': [0x00, 0x08, 0x10, 0x18, 0x20, 0x28, 0x30, 0x38],
        'charincrement': 0x40
    }
    contents = gfx_rebuilder.reencode_gfx(contents, sprint2_tile_layout)
    chunks = transforms.deinterleave_nibble(contents, 2)
    zip_files['6396-01.p4'] = chunks[0]
    zip_files['6397-01.r4'] = chunks[1]

    zip_files['6400-01.m2'] = bytearray(0x100)
    zip_files['6401-01.e2'] = bytearray(0x20)

    return [{
        'filename': game_desc['filename'],
        'contents': helpers.build_zip(zip_files)
    }]

# def _handle_canyon(in_files, game_desc):
#     zip_files = {}

#     # CanyonBomber.bin
#     contents = in_files.get('CanyonBomber.bin')['contents']
#     chunks = transforms.custom_split(contents, [1024, 1024, 2048])
#     nib_chunks = transforms.deinterleave_nibble(chunks[0], 2)
#     zip_files['9503-01.p1'] = nib_chunks[0]
#     zip_files['9499-01.j1'] = nib_chunks[1]
#     zip_files['9496-01.d1'] = chunks[2]

#     # CanyonBomberSprites.bmp
#     contents = in_files.get('CanyonBomberSprites.bmp')['contents']
#     contents = gfx_rebuilder.reverse_bmp(contents)
#     _SPRITE_LAYOUT = {
#         'width': 32,
#         'height': 16,
#         'total': 4,
#         'planes': 1,
#         'planeoffset': [0],
#         'xoffset': [0x007, 0x006, 0x005, 0x004, 0x003, 0x002, 0x001, 0x000,
#             0x00F, 0x00E, 0x00D, 0x00C, 0x00B, 0x00A, 0x009, 0x008,
#             0x107, 0x106, 0x105, 0x104, 0x103, 0x102, 0x101, 0x100,
#             0x10F, 0x10E, 0x10D, 0x10C, 0x10B, 0x10A, 0x109, 0x108],
#         'yoffset': [0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0x60, 0x70,
#             0x80, 0x90, 0xA0, 0xB0, 0xC0, 0xD0, 0xE0, 0xF0],
#         'charincrement': 0x200
#     }
#     contents = gfx_rebuilder.reencode_gfx(contents, _SPRITE_LAYOUT)
#     chunks = transforms.deinterleave_nibble(contents, 2)
#     zip_files['9505-01.n5'] = chunks[0]
#     zip_files['9506-01.m5'] = chunks[1]

#     # CanyonBomberTiles.bmp
#     # Something is wrong with the math on this one,
#     # and I think it's a discrepancy vs. the mame tile layout
#     contents = in_files.get('CanyonBomberTiles.bmp')['contents']
#     contents = gfx_rebuilder.reverse_bmp(contents)
#     zip_files['tiles_temp'] = contents
#     _TILE_LAYOUT = {
#         'width': 8,
#         'height': 8,
#         'total': 64,
#         'planes': 1,
#         'planeoffset': [0],
#         'xoffset': [4, 5, 6, 7, 0, 1, 2, 3],
#         'yoffset': [0x00, 0x08, 0x10, 0x18, 0x20, 0x28, 0x30, 0x38],
#         'charincrement': 0x80
#     }
#     contents = gfx_rebuilder.reencode_gfx(contents, _TILE_LAYOUT)
#     zip_files['9492-01.n8'] = contents

#     return [{
#         'filename': f"{game_desc['mame_name']}.zip",
#         'contents': helpers.build_zip(zip_files)
#     }]

def _handle_avalnche(in_files, game_desc):
    # Avalanche.bin
    contents = in_files.get('Avalanche.bin')['contents']
    chunks = transforms.equal_split(contents, 3)
    new_chunks = []
    for chunk in chunks:
        new_chunks.extend(transforms.deinterleave_nibble(chunk, 2))

    filenames = [
        "30612.d2",
        "30615.d3",
        "30613.e2",
        "30616.e3",
        "30611.c2",
        "30614.c3",
    ]
    zip_files = dict(zip(filenames, new_chunks))

    return [{
        'filename': game_desc['filename'],
        'contents': helpers.build_zip(zip_files)
    }]

def _handle_standard(in_files, game_desc):
    out_files = []

    contents = list(in_files.values())[0]['contents']

    # Truncate
    if 'truncate' in game_desc:
        contents = transforms.truncate(contents, game_desc['truncate'])

    # Split
    split_data = game_desc['split']
    if isinstance(split_data, list):
        chunks = transforms.equal_split(contents, len(game_desc['split']))
        filenames = split_data
    elif isinstance(split_data, dict):
        filenames = list(split_data.keys())
        sizes = list(split_data.values())
        chunks = transforms.custom_split(contents, sizes)
    else:
        logger.error("Invalid split type!")
        return []
    out_files.append({
        'filename': game_desc['filename'],
        'contents': helpers.build_zip(dict(zip(filenames, chunks)))
    })

    return out_files

def extract(in_dir, out_dir, class_ref, include_partials):
    '''Extract Atari Arcade ROMs'''
    funcs = globals()
    for game in class_ref._metadata['out']['files']:
        if game['set'] == 'Arcade' and game['status'] != 'no-rom':
            is_partial = game['status'] == 'partial'
            if is_partial and not include_partials:
                continue

            logger.info(f"Reading {game['game']}...")

            # Read the input files
            in_files = {}
            for in_file_ref in game['in_files']:
                file_metadata = class_ref._metadata['in']['files'].get(in_file_ref)
                datafile = class_ref.read_datafile(in_dir, file_metadata)
                in_files[datafile['name']] = {
                    "contents": datafile['contents'],
                    "metadata": file_metadata
                }

            if is_partial:
                logger.info(f"Copying partially extracted {game['game']}...")

                if 'handler' in game:
                    handler_func = funcs[game['handler']]
                    output_files = handler_func(in_files, game)
                else:
                    zip_files = {}
                    for filename in game['in_files']:
                        zip_files[filename] = in_files[filename]['contents']

                    output_files = [{
                        'filename': game['filename'],
                        'contents': helpers.build_zip(zip_files)
                    }]
            else:
                logger.info(f"Extracting {game['game']}...")
                if 'handler' in game:
                    handler_func = funcs[game['handler']]
                else:
                    handler_func = _handle_standard
                output_files = handler_func(in_files, game)
            
            for output_file in output_files:
                filename = output_file['filename']
                logger.info(f"Writing {game['game']} to {filename}...")
                _ = class_ref.verify_out_file(filename, output_file['contents'])
                out_path = os.path.join(out_dir, filename)
                with open(out_path, "wb") as out_file:
                    out_file.write(output_file['contents'])



'''Implementation of mmbnlc1: Mega Man Battle Network Legacy Collection 1'''
import logging
import os
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers

logger = logging.getLogger('gextoolbox')

class MegaManLegacyCollection1Task(BaseTask):
    '''Implements mmbnlc1: Mega Man Battle Network Legacy Collection 1'''
    _task_name = "mmbnlc1"
    _title = "Mega Man Battle Network Legacy Collection 1"
    _details_markdown = '''
Based on zip file discoveries from https://github.com/farmerbb/RED-Project/issues/101
'''
    _default_input_folder = helpers.gen_steam_app_default_folder("MegaMan_BattleNetwork_LegacyCollection_Vol1")
    _input_folder_desc = "'MegaMan_BattleNetwork_LegacyCollection_Vol1' Folder (Steam MMBNLC1 install folder)"

    def execute(self, in_dir, out_dir):
        bundle_contents = {}
        for file_ref, file_metadata in self._metadata['in']['files'].items():
            resolved_file = self.read_datafile(in_dir, file_metadata)
            bundle_contents[file_ref] = resolved_file['contents']

        out_files = []

        out_files.extend(
            helpers.pull_files_from_archive(bundle_contents['exe1.dat'], "zip", {
                "MegaMan Battle Network.gba": "exe1/rom_e.srl",
                "RockMan.EXE.gba": "exe1/rom.srl"
            })
        )
        out_files.extend(
            helpers.pull_files_from_archive(bundle_contents['exe2j.dat'], "zip", {
                "MegaMan Battle Network 2.gba": "exe2j/rom_e.srl",
                "RockMan.EXE 2.gba": "exe2j/rom.srl"
            })
        )
        out_files.extend(
            helpers.pull_files_from_archive(bundle_contents['exe3.dat'], "zip", {
                "MegaMan Battle Network 3 White.gba": "exe3/rom_e.srl",
                "RockMan.EXE 3.gba": "exe3/rom.srl"
            })
        )
        out_files.extend(
            helpers.pull_files_from_archive(bundle_contents['exe3b.dat'], "zip", {
                "MegaMan Battle Network 3 Blue.gba": "exe3b/rom_b_e.srl",
                "RockMan.EXE 3 Black.gba": "exe3b/rom_b.srl"
            })
        )

        for out_file_entry in out_files:
            filename = out_file_entry['filename']
            _ = self.verify_out_file(filename, out_file_entry ['contents'])
            out_path = os.path.join(out_dir, filename)
            with open(out_path, "wb") as out_file:
                logger.info(f"Writing {filename}...")
                out_file.write(out_file_entry['contents'])

        logger.info("Processing complete.")


'''Implementation of pacmanmplus: Pac Man Museum Plus'''
import logging
import os
import lzma
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers
from gex.lib.utils.blob import transforms

logger = logging.getLogger('gextoolbox')

class PacManMuseumPlusTask(BaseTask):
    '''Implements pacmanmplus: Pac Man Museum Plus'''
    _task_name = "pacmanmplus"
    _title = "Pac Man Museum Plus"
    _details_markdown = ''''''
    _default_input_folder = helpers.gen_steam_app_default_folder("PAC-MAN MUSEUM PLUS")
    _input_folder_desc = "Pac Man Museum Plus install folder"

    _prop_info = {
        "include-partials": {
            "description": "Include the partial ROMs that are missing data; useful for mixing with other sources or investigation",
            "default": False,
            "type": "Boolean"
        }
    }

    # PAC-IN-TIME
    # Official ROM is 1,048,576
    # DLL is 1,062,400
    # So, it really does need to be compressed!
    # binwalk -e -M doesn't find anything

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }


    def execute(self, in_dir, out_dir):
        for game in self._metadata['out']['files']:
            if game.get('status') == "no-rom":
                logger.info(f"Skipping {game['game']} as there is no ROM to extract...")
                continue

            is_partial = game.get('status') == "partial"
            if not self._props.get('include-partials') and is_partial:
                logger.info(f"Skipping {game['game']} as this tool cannot extract a working copy...")
                continue

            # read the matching input file
            pkg_name = game['extract']['in_file']
            in_file_entry = self._metadata['in']['files'][pkg_name]
            contents = self.read_datafile(in_dir, in_file_entry)['contents']

            logger.info(f"Extracting {game['game']}...")
            zip_files = {}

            contents = transforms.cut(contents, game['extract']['start'], length=game['extract']['length'])
            if game['system'] == "Arcade":
                lzd = lzma.LZMADecompressor()
                contents = lzd.decompress(contents)
                if is_partial:
                    zip_files["decompressed_blob"] = contents

                work_files = game.get('zip_files') or {}
                for work_filename, geometry in work_files.items():
                    start = int(geometry['start'], 16)
                    length = int(geometry['length'], 16)
                    zip_files[work_filename] = transforms.cut(contents, start, length=length)
            else:
                if is_partial:
                    zip_files["blob"] = contents

                for work_file in game.get('out') or []:
                    zip_files[work_file['filename']] = transforms.cut(contents, work_file['start'], length=work_file['length'])

            game_data = helpers.build_zip(zip_files)
            filename = f"partial_{game['filename']}" if is_partial else game['filename']
            if not is_partial:
                _ = self.verify_out_file(filename, game_data)
            else:
                logger.info(f"Skipping verification for partial extract {filename}.")
            out_path = os.path.join(out_dir, filename)
            with open(out_path, "wb") as out_file:
                out_file.write(game_data)

        logger.info("Processing complete.")



import click

from .details import details
from .extract import extract
from .list import list_cli

@click.group()
def tasks():
    """Tasks are scripts to extract ROMs from various PC Releases (Steam, GOG, etc.)"""

tasks.add_command(details)
tasks.add_command(extract)
tasks.add_command(list_cli)



'''
Wrapper for python-magic that adds support for game extraction specific file types
'''
import magic

# Internally tracking these in https://docs.google.com/spreadsheets/d/1eMI9EIzopg1J9UZsMFOx5lLSMPsf7EYxovLM5VFoUjU/edit#gid=0

KPKA = "KPKA Archive, Capcom RE Engine"
IBIS = "IBIS Archive, Capcom ROM Releases"
ARC = "ARC Archive, Capcom MT Framework"
M2 = "M2 Archive, M2 (Developer/Publisher)"

def enhanced_magic_from_path(in_file):
    '''Use libmagic and enhanced_look to id a file's contents'''
    magic_id = None
    try:
        magic_id = magic.from_file(in_file)
    except Exception as error:
        print(error)

    if magic_id != 'data' and magic_id is not None:
        return magic_id

    # Try to identify based on known file types magic bytes
    with open(in_file, 'rb') as file:
        content_start = file.read(2048)
        return enhanced_look(content_start)

def enhanced_magic_from_buffer(content_peek):
    '''Use libmagic and enhanced_look to id buffer content'''
    magic_id = None
    try:
        magic_id = magic.from_buffer(content_peek)
    except Exception as error:
        print(error)

    if magic_id != 'data' and magic_id is not None:
        return magic_id

    # Try to identify based on known file types magic bytes
    return enhanced_look(content_peek)

def enhanced_look(content_peek):
    '''Magic-like function that identifies uncommon relevant filetypes'''
    if content_peek[0:4] == b"KPKA":
        return KPKA
    if content_peek[0:4] == b"IBIS":
        return IBIS
    if content_peek[0:3] == b"ARC":
        return ARC
    if content_peek[0:3] == b"mdf":
        return M2
    return 'data'

_PK_HEADER = "PK".encode('utf-8')
def check_if_zip(contents):
    '''Simple ZIP file check that searches for a 'PK' ZIP header'''
    return contents[0:2] == _PK_HEADER



'''Implementation of zamn: Zombies Ate My Neighbors and Ghoul Patrol'''
import logging
import os
from gex.lib.tasks.basetask import BaseTask
from gex.lib.utils.blob import hash as hash_helper

logger = logging.getLogger('gextoolbox')

class ZAMNTask(BaseTask):
    '''Implements zamn: Zombies Ate My Neighbors and Ghoul Patrol'''
    _task_name = "zamn"
    _title = "Zombies Ate My Neighbors and Ghoul Patrol"
    _details_markdown = '''
Based on: https://www.gog.com/forum/general/rom_based_gog_games_compatible_with_third_party_emulators_thread/post127

These ROMs are pulled out of the main executable. 
    '''
    _default_input_folder = r"C:\Program Files (x86)\GOG Galaxy\Games\Zombies Ate My Neighbour & Ghoul Patrol"
    _input_folder_desc = "ZAMN folder"

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        zamngp_file = self.read_all_datafiles(in_dir).get("zamngp")
        exe_data = zamngp_file['contents']
        file_ver = zamngp_file['version']
        if file_ver is not None:
            logger.info(f"Found {file_ver}...")

        for game in self._metadata['out']['files']:
            game_data = None
            game_header = bytes.fromhex(game['extract']['header_bytes'])
            target_size = game['extract']['length']

            if file_ver is not None:
                logger.info(f"Pulling {game['game']} from known offset...")
                version_info = game['extract']['versions'][file_ver]
                start = int(version_info['start'], 16)
                game_data = exe_data[start:start+target_size]
            else:
                logger.info(f"Finding {game['game']}...")

                # Try check_offsets
                logger.info(f"Finding {game['game']} via known offsets...")
                for version_info in game['extract']['versions'].values():
                    check_offset = int(version_info['start'], 16)
                    check_header = exe_data[check_offset:check_offset+10]
                    if check_header == game_header:
                        logger.info(f"Found header at {hex(check_offset)}!")
                        game_data = exe_data[check_offset:check_offset+target_size]
                        crc = hash_helper.get_crc(game_data)[2:].upper().rjust(8, "0")
                        if crc == game['verify']['crc']:
                            logger.info(f"Found crc match at {hex(check_offset)}!")
                            break
                        else:
                            logger.info(f"False header match at {hex(check_offset)}!")
                            game_data = None

                # Do the manual search
                if not game_data:
                    logger.info(f"Finding {game['game']} via full search - please wait...")
                    for offset in range(0, len(exe_data) - target_size):
                        check_header = exe_data[offset:offset+10]
                        if check_header == game_header:
                            logger.info(f"Found header at {hex(offset)}!")
                            game_data = exe_data[offset:offset+target_size]
                            crc = hash_helper.get_crc(game_data)[2:].upper().rjust(8, "0")
                            if crc == game['verify']['crc']:
                                logger.info(f"Found crc match at {hex(offset)}!")
                                break
                            else:
                                logger.info(f"False header match at {hex(offset)}!")
                                game_data = None

            if game_data:
                _ = self.verify_out_file(game['filename'], game_data)
                logger.info(f"Saving {game['filename']}...")
                with open(os.path.join(out_dir, game['filename']), "wb") as out_file:
                    out_file.write(game_data)
            else:
                logger.warning(f"Game {game['game']} not found!")

        logger.info("Processing complete.")



'''Implementation of acac: Arcade Classics Anniversary Collection'''
import logging
import lzma
import os
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers
from gex.lib.utils.blob import transforms

logger = logging.getLogger('gextoolbox')

# Implementation notes:
# - This one is a very imprecise process. While it initally looks like there's one internal LZMA blob per title,
#   this doesn't appear to actually be the case.



class ACACTask(BaseTask):
    '''Implements acac: Arcade Classics Anniversary Collection'''
    _task_name = "acac"
    _title = "Arcade Classics Anniversary Collection"
    _details_markdown = ''''''
    _default_input_folder = helpers.gen_steam_app_default_folder("Arcade Classics Anniversary Collection")
    _input_folder_desc = "Arcade Classics Anniversary Collection install folder"

    _prop_info = {}

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        src_file = self.read_all_datafiles(in_dir).get("source")
        src_contents = src_file['contents']
        file_ver = src_file['version']

        # src_contents_decomp = self.lzma_multi_decomp(src_contents, out_dir)

        # with open(os.path.join(out_dir, 'merged_decomp_blob'), "wb") as out_file:
        #     out_file.write(src_contents_decomp)

        out_files = []
        out_files.extend(self._handle_nemesis(src_contents, file_ver))
        out_files.extend(self._handle_hcastle(src_contents, file_ver))
        out_files.extend(self._handle_ajax(src_contents, file_ver))
        out_files.extend(self._handle_vulcan(src_contents, file_ver))
        out_files.extend(self._handle_thunderx(src_contents, file_ver))
        out_files.extend(self._handle_salamand(src_contents, file_ver))
        out_files.extend(self._handle_twinbee(src_contents, file_ver))

        if out_files:
            for out_file_entry in out_files:
                filename = out_file_entry['filename']
                _ = self.verify_out_file(filename, out_file_entry['contents'])
                out_path = os.path.join(out_dir, filename)
                with open(out_path, "wb") as out_file:
                    logger.info(f"Writing {out_file_entry['filename']}...")
                    out_file.write(out_file_entry['contents'])

        logger.info("Processing complete.")

    def _get_extract_metadata(self, game_name):
        for game in self._metadata['out']['files']:
            if game['game'] == game_name:
                return game['extract']

        return None

    def _handle_ajax(self, contents, version):
        extract_metadata = self._get_extract_metadata('Typhoon')
        contents = transforms.cut(contents, 
            int(extract_metadata['versions'][version]['start'], 0),
            length=extract_metadata['length'])
        lzd = lzma.LZMADecompressor()
        contents = lzd.decompress(contents)
        out_files = []
        
        # AJAX Common
        func_map = {}
        def k_roms(contents):
            out_chunks = {}
            out_chunks["770c13.n22"] = transforms.cut(contents, 0x090000, length = 0x40000)
            out_chunks["770c12.k22"] = transforms.cut(contents, 0x0D0000, length = 0x40000)
            out_chunks["770c09.n4"] = transforms.cut(contents, 0x110000, length = 0x80000)
            out_chunks["770c08.k4"] = transforms.cut(contents, 0x190000, length = 0x80000)
            out_chunks["770c06.f4"] = transforms.cut(contents, 0x210000, length = 0x40000)
            out_chunks["770c07.h4"] = transforms.cut(contents, 0x250000, length = 0x40000)
            out_chunks["770c10"] = transforms.cut(contents, 0x290000, length = 0x40000)
            out_chunks["770c11"] = transforms.cut(contents, 0x2D0000, length = 0x80000)
            out_chunks["63s241.j11"] = transforms.cut(contents, 0x350000, length = 0x200)
            return out_chunks
        func_map['k_roms'] = k_roms
        common_file_map = helpers.process_rom_files(contents, func_map)

        # AJAXJ
        func_map = {}
        def ajaxj_maincpu(contents):
            out_chunks = {}
            out_chunks["770_l01.n11"] = transforms.cut(contents, 0x000000, length = 0x10000)
            out_chunks["770_l02.n12"] = transforms.cut(contents, 0x020000, length = 0x10000)
            out_chunks["770_l05.i16"] = transforms.cut(contents, 0x030000, length = 0x8000)
            out_chunks["770_f04.g16"] = transforms.cut(contents, 0x038000, length = 0x10000)
            out_chunks["770_f03.f16"] = transforms.cut(contents, 0x048000, length = 0x8000)
            return out_chunks
        func_map['maincpu'] = ajaxj_maincpu
        def ajaxj_krom(contents):
            out_chunks = {}
            out_chunks["770c13.n22"] = transforms.cut(contents, 0x090000, length = 0x40000)
            out_chunks["770c12.k22"] = transforms.cut(contents, 0x0D0000, length = 0x40000)
            return out_chunks
        func_map['k_roms'] = ajaxj_krom
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        out_files.append({'filename': 'ajaxj.zip', 'contents': helpers.build_rom(contents, func_map)})

        # Typhoon
        func_map = {}
        def typhoon_maincpu(contents):
            out_chunks = {}
            out_chunks["770_k01.n11"] = transforms.cut(contents, 0x050000, length = 0x10000)
            out_chunks["770_k02.n12"] = transforms.cut(contents, 0x060000, length = 0x10000)
            out_chunks["770_k05.i16"] = transforms.cut(contents, 0x070000, length = 0x8000)
            out_chunks["770_f04.g16"] = transforms.cut(contents, 0x038000, length = 0x10000)
            out_chunks["770_h03.f16"] = transforms.cut(contents, 0x048000, length = 0x8000)
            return out_chunks
        func_map['maincpu'] = typhoon_maincpu
        func_map['k_roms'] = ajaxj_krom
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        out_files.append({'filename': 'typhoon.zip', 'contents': helpers.build_rom(contents, func_map)})
        
        return out_files

    def _handle_nemesis(self, contents, version):
        extract_metadata = self._get_extract_metadata('Nemesis')
        contents = transforms.cut(contents, 
            int(extract_metadata['versions'][version]['start'], 0),
            length=extract_metadata['length'])
        lzd = lzma.LZMADecompressor()
        contents = lzd.decompress(contents)
        out_files = []

        # Nemesis Common
        func_map = {}
        def k_rom(contents):
            filenames = [
                "400-a01.fse",
                "400-a02.fse"
            ]
            contents = transforms.cut(contents, 0x84000, length = 0x200)
            chunks = transforms.equal_split(contents, len(filenames))
            return dict(zip(filenames, chunks))
        func_map['k005289'] = k_rom
        common_file_map = helpers.process_rom_files(contents, func_map)

        # Nemesis
        func_map = {}
        def nemesis_maincpu(contents):
            filenames = [
                "456-d01.12a",
                "456-d05.12c",
                "456-d02.13a",
                "456-d06.13c",
                "456-d03.14a",
                "456-d07.14c",
                "456-d04.15a",
                "456-d08.15c",
            ]
            contents = transforms.cut(contents, 0x40000, length = 0x40000)
            chunks = transforms.equal_split(contents, len(filenames) // 2)
            chunks = transforms.deinterleave_all(chunks, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['maincpu'] = nemesis_maincpu
        def nemesis_audiocpu(contents):
            contents = transforms.cut(contents, 0x80000, length = 0x4000)
            return {'456-d09.9c': contents}
        func_map['audiocpu'] = nemesis_audiocpu
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        out_files.append({'filename': 'nemesis.zip', 'contents': helpers.build_rom(contents, func_map)})

        # Nemesis UK/World?
        func_map = {}
        def nemesisuk_maincpu(contents):
            filenames = [
                "456-e01.12a",
                "456-e05.12c",
                "456-e02.13a",
                "456-e06.13c",
                "456-e03.14a",
                "456-e07.14c",
                "456-e04.15a",
                "456-e08.15c",
            ]
            contents = transforms.cut(contents, 0x00000, length = 0x40000)
            chunks = transforms.equal_split(contents, len(filenames) // 2)
            chunks = transforms.deinterleave_all(chunks, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['maincpu'] = nemesisuk_maincpu
        def nemesisuk_audiocpu(contents):
            contents = transforms.cut(contents, 0x80000, length = 0x4000)
            return {'456-b09.9c': contents}
        func_map['audiocpu'] = nemesisuk_audiocpu
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        out_files.append({'filename': 'nemesisuk.zip', 'contents': helpers.build_rom(contents, func_map)})

        return out_files


    def _handle_hcastle(self, contents, version):
        extract_metadata = self._get_extract_metadata('Haunted Castle (Version M)')
        contents = transforms.cut(contents, 
            int(extract_metadata['versions'][version]['start'], 0),
            length=extract_metadata['length'])
        lzd = lzma.LZMADecompressor()
        contents = lzd.decompress(contents)

        out_files = []

        # HCastle Common
        func_map = {}
        def k_roms(contents):
            out_chunks = {}
            out_chunks["768c07.e17"] = transforms.cut(contents, 0x0A8000, length = 0x80000)
            out_chunks["768c09.g21"] = transforms.cut(contents, 0x128000, length = 0x80000)
            out_chunks["768c08.g19"] = transforms.cut(contents, 0x1A8000, length = 0x80000)
            out_chunks["768c04.j5"] = transforms.cut(contents, 0x228000, length = 0x80000)
            out_chunks["768c05.j6"] = transforms.cut(contents, 0x2A8000, length = 0x80000)
            return out_chunks
        func_map['k_roms'] = k_roms
        def proms(contents):
            out_chunks = {}
            out_chunks["768c13.j21"] = transforms.cut(contents, 0x328000, length = 0x100)
            out_chunks["768c11.i4"] = out_chunks["768c13.j21"]
            out_chunks["768c14.j22"] = transforms.cut(contents, 0x328100, length = 0x100)
            out_chunks["768c10.i3"] = out_chunks["768c14.j22"]
            out_chunks["768b12.d20"] = transforms.cut(contents, 0x328400, length = 0x100)
            return out_chunks
        func_map['proms'] = proms
        def audiocpu(contents):
            out_chunks = {}
            out_chunks["768e01.e4"] = transforms.cut(contents, 0x0A0000, length = 0x8000)
            return out_chunks
        func_map['audiocpu'] = audiocpu
        common_file_map = helpers.process_rom_files(contents, func_map)

        # HCASTLE
        func_map = {}
        def hcastle_maincpu(contents):
            out_chunks = {}
            out_chunks["m03.k12"] = transforms.cut(contents, 0x078000, length = 0x8000)
            out_chunks["b06.k8"] = transforms.cut(contents, 0x080000, length = 0x20000)
            return out_chunks
        func_map['maincpu'] = hcastle_maincpu
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        out_files.append({'filename': 'hcastle.zip', 'contents': helpers.build_rom(contents, func_map)})

        # HCASTLEE
        func_map = {}
        def hcastlee_maincpu(contents):
            out_chunks = {}
            out_chunks["768e03.k12"] = transforms.cut(contents, 0x028000, length = 0x8000)
            out_chunks["768e06.k8"] = transforms.cut(contents, 0x030000, length = 0x20000)
            return out_chunks
        func_map['maincpu'] = hcastlee_maincpu
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        out_files.append({'filename': 'hcastlee.zip', 'contents': helpers.build_rom(contents, func_map)})

        # HCASTLEK
        func_map = {}
        def hcastlek_maincpu(contents):
            out_chunks = {}
            out_chunks["768k03.k12"] = transforms.cut(contents, 0x050000, length = 0x8000)
            out_chunks["768g06.k8"] = transforms.cut(contents, 0x058000, length = 0x20000)
            return out_chunks
        func_map['maincpu'] = hcastlek_maincpu
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        out_files.append({'filename': 'hcastlek.zip', 'contents': helpers.build_rom(contents, func_map)})

        # akumajoun
        func_map = {}
        def akumajoun_maincpu(contents):
            out_chunks = {}
            out_chunks["768n03.k12"] = transforms.cut(contents, 0x000000, length = 0x8000)
            out_chunks["768j06.k8"] = transforms.cut(contents, 0x008000, length = 0x20000)
            return out_chunks
        func_map['maincpu'] = akumajoun_maincpu
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        out_files.append({'filename': 'akumajoun.zip', 'contents': helpers.build_rom(contents, func_map)})

        return out_files

    def _handle_vulcan(self, contents, version):
        extract_metadata = self._get_extract_metadata('Vulcan Venture')
        contents = transforms.cut(contents, 
            int(extract_metadata['versions'][version]['start'], 0),
            length=extract_metadata['length'])
        lzd = lzma.LZMADecompressor()
        contents = lzd.decompress(contents)
        out_files = []

        # 0x000000   0x040000     maincpu - gradius2
        # 0x040000   0x040000     maincpu - vulcan
        # 0x080000   0x040000     sub
        # 0x0C0000   0x008000     audiocpu
        # 0x0C8000   0x020000     k_rom
        # 0x0E8000   0x004000     fixed - gradius2
        # 0x0EB000   0x004000     fixed - vulcan
        # 0x0F0000   0x200000     gfx
        
        # 785_f02.7c MISSING 0x20000

        # VULCAN Common
        func_map = {}
        def audiocpu(contents):
            out_chunks = {}
            out_chunks["785_g03.10a"] = transforms.cut(contents, 0x0C0000, length = 0x8000)
            return out_chunks
        func_map['audiocpu'] = audiocpu
        def gfx(contents):
            filenames = [
                "785f15.p13",
                "785f17.p16",
                "785f16.p15",
                "785f18.p18"
            ]
            contents = transforms.cut(contents, 0x0F0000, length = 0x200000)
            chunks = transforms.equal_split(contents, 2)
            chunks = transforms.deinterleave_all(chunks, 2, 2)
            chunks = transforms.swap_endian_all(chunks)
            return dict(zip(filenames, chunks))
        func_map['gfx'] = gfx
        def sub(contents):
            filenames = [
                "785_p07.10n",
                "785_p06.8n",
                "785_p13.10s",
                "785_p12.8s",
            ]
            contents = transforms.cut(contents, 0x080000, length = 0x040000)
            chunks = transforms.equal_split(contents, 2)
            chunks = transforms.deinterleave_all(chunks, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['sub'] = sub
        def k_rom(contents):
            out_chunks = {}
            out_chunks["785_f01.5a"] = transforms.cut(contents, 0x0C8000, length = 0x20000)
            return out_chunks
        func_map['k_rom'] = k_rom
        ph_files = {
            "785_f02.7c": 0x20000
        }
        func_map['ph'] = helpers.placeholder_helper(ph_files)
        common_file_map = helpers.process_rom_files(contents, func_map)

        # GRADIUS2
        func_map = {}
        #maincpu
        def gradius2_maincpu(contents):
            filenames = [
                "785_x05.6n",
                "785_x04.4n",
                "785_x09.6r",
                "785_x08.4r",
            ]
            contents = transforms.cut(contents, 0x00000, length = 0x40000)
            chunks = transforms.equal_split(contents, len(filenames) // 2)
            chunks = transforms.deinterleave_all(chunks, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['maincpu'] = gradius2_maincpu
        def fixed(contents):
            out_chunks = {}
            out_chunks["785_g14.d8"] = transforms.cut(contents, 0x0E8000, length = 0x4000)
            return out_chunks
        func_map['fixed'] = fixed
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        out_files.append({'filename': 'gradius2.zip', 'contents': helpers.build_rom(contents, func_map)})

        
        # VULCAN
        func_map = {}
        def vulcan_maincpu(contents):
            filenames = [
                "785_w05.6n",
                "785_w04.4n",
                "785_w09.6r",
                "785_w08.4r",
            ]
            contents = transforms.cut(contents, 0x40000, length = 0x40000)
            chunks = transforms.equal_split(contents, len(filenames) // 2)
            chunks = transforms.deinterleave_all(chunks, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['maincpu'] = vulcan_maincpu
        def vulcan_fixed(contents):
            out_chunks = {}
            out_chunks["785_h14.d8"] = transforms.cut(contents, 0x0EC000, length = 0x4000)
            return out_chunks
        func_map['fixed'] = vulcan_fixed
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        out_files.append({'filename': 'vulcan.zip', 'contents': helpers.build_rom(contents, func_map)})

        return out_files



    def _handle_thunderx(self, contents, version):
        extract_metadata = self._get_extract_metadata('Thunder Cross (Set 1)')
        contents = transforms.cut(contents, 
            int(extract_metadata['versions'][version]['start'], 0),
            length=extract_metadata['length'])
        contents = transforms.cut(contents, 0x4D13D0, length=455549)
        lzd = lzma.LZMADecompressor()
        contents = lzd.decompress(contents)
        out_files = []

        # 0x000000 0x020000     maincpu - thunderxj
        # 0x020000 0x008000     audiocpu -thunderxj/xb/x
        # 0x028000 0x020000     maincpu - thunderxb
        # 0x048000 0x020000     maincpu - thunderx
        # 0x068000 0x050000     mystery junk data
        #    BK header kinda looks like maincpu, but no idea what it could match
        #    We have every thunderx variant MAME knows, and the other games for this driver are very different
        # 0x048000 0x020000     maincpu - thunderxa
        # 0x0C8000 0c008000     audiocpu -thunderxa
        # 0x0D0000 0x080000     tiles
        # 0x150000 0x080000     sprites
        # 0x1D0000 0x000100     prom

        # thunderx common
        func_map = {}
        def thunderx_common_proms(contents):
            out_chunks = {}
            out_chunks["873a08.f20"] = transforms.cut(contents, 0x1D0000, length = 0x100)
            return out_chunks
        func_map['proms'] = thunderx_common_proms
        def thunderx_common_sprites(contents):
            sets = [
                {
                    'filenames': [
                        "873c04a.f11",
                        "873c04c.f10"
                    ],
                    'start': 0x150000,
                    'length': 0x20000
                },
                {
                    'filenames': [
                        "873c05a.f9",
                        "873c05c.f8"
                    ],
                    'start': 0x190000,
                    'length': 0x20000
                },
                {
                    'filenames': [
                        "873c04b.e11",
                        "873c04d.e10"
                    ],
                    'start': 0x170000,
                    'length': 0x20000
                },
                {
                    'filenames': [
                        "873c05b.e9",
                        "873c05d.e8"
                    ],
                    'start': 0x1B0000,
                    'length': 0x20000
                },

            ]
            out_tiles = {}
            for curr_set in sets:
                temp_contents = transforms.cut(contents, curr_set['start'], length = curr_set['length'])
                chunks = transforms.deinterleave(temp_contents, 2, 1)
                out_tiles.update(dict(zip(curr_set['filenames'], chunks)))
            return out_tiles
        func_map['sprites'] = thunderx_common_sprites
        
        def thunderx_common_tiles(contents):
            sets = [
                {
                    'filenames': [
                        "873c06a.f6",
                        "873c06c.f5"
                    ],
                    'start': 0x0D0000,
                    'length': 0x20000
                },
                {
                    'filenames': [
                        "873c07a.f4",
                        "873c07c.f3"
                    ],
                    'start': 0x110000,
                    'length': 0x20000
                },
                {
                    'filenames': [
                        "873c06b.e6",
                        "873c06d.e5",
                    ],
                    'start': 0x0F0000,
                    'length': 0x20000
                },
                {
                    'filenames': [
                        "873c07b.e4",
                        "873c07d.e3",
                    ],
                    'start': 0x130000,
                    'length': 0x20000
                },

            ]
            out_tiles = {}
            for curr_set in sets:
                temp_contents = transforms.cut(contents, curr_set['start'], length = curr_set['length'])
                chunks = transforms.deinterleave(temp_contents, 2, 1)
                out_tiles.update(dict(zip(curr_set['filenames'], chunks)))
            return out_tiles
        func_map['tiles'] = thunderx_common_tiles
        common_file_map = helpers.process_rom_files(contents, func_map)

        # thunderx common audio
        func_map = {}
        def thunderx_common_audiocpu(contents):
            contents = transforms.cut(contents, 0x20000, length = 0x8000)
            return {"873-f01.f8": contents}
        func_map['audiocpu'] = thunderx_common_audiocpu
        common_audio_file_map = helpers.process_rom_files(contents, func_map)

        # thunderxj
        func_map = {}
        def thunderxj_maincpu(contents):
            filenames = [
                "873-n02.k13",
                "873-n03.k15"
            ]
            contents = transforms.cut(contents, 0x00000, length = 0x20000)
            chunks = transforms.equal_split(contents, 2)
            return dict(zip(filenames, chunks))
        func_map['maincpu'] = thunderxj_maincpu
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        func_map['audio'] = helpers.existing_files_helper(common_audio_file_map)
        out_files.append({'filename': 'thunderxj.zip', 'contents': helpers.build_rom(contents, func_map)})

        # thunderxb
        func_map = {}
        def thunderxb_maincpu(contents):
            filenames = [
                "873-02.k13",
                "873-03.k15"
            ]
            contents = transforms.cut(contents, 0x28000, length = 0x20000)
            chunks = transforms.equal_split(contents, 2)
            return dict(zip(filenames, chunks))
        func_map['maincpu'] = thunderxb_maincpu
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        func_map['audio'] = helpers.existing_files_helper(common_audio_file_map)
        out_files.append({'filename': 'thunderxb.zip', 'contents': helpers.build_rom(contents, func_map)})

        # thunderx
        func_map = {}
        def thunderx_maincpu(contents):
            filenames = [
                "873-s02.k13",
                "873-s03.k15"
            ]
            contents = transforms.cut(contents, 0x48000, length = 0x20000)
            chunks = transforms.equal_split(contents, 2)
            return dict(zip(filenames, chunks))
        func_map['maincpu'] = thunderx_maincpu
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        func_map['audio'] = helpers.existing_files_helper(common_audio_file_map)
        out_files.append({'filename': 'thunderx.zip', 'contents': helpers.build_rom(contents, func_map)})

        # thunderxa
        func_map = {}
        def thunderxa_maincpu(contents):
            filenames = [
                "873-k02.k13",
                "873-k03.k15"
            ]
            contents = transforms.cut(contents, 0xA8000, length = 0x20000)
            chunks = transforms.equal_split(contents, 2)
            return dict(zip(filenames, chunks))
        func_map['maincpu'] = thunderxa_maincpu
        def thunderxa_audiocpu(contents):
            contents = transforms.cut(contents, 0xC8000, length = 0x8000)
            return {"873-h01.f8": contents}
        func_map['audiocpu'] = thunderxa_audiocpu
        func_map['common'] = helpers.existing_files_helper(common_file_map)
        out_files.append({'filename': 'thunderxa.zip', 'contents': helpers.build_rom(contents, func_map)})

        return out_files


    def _handle_salamand(self, contents, version):
        extract_metadata = self._get_extract_metadata('Salamander')
        contents = transforms.cut(contents, 
            int(extract_metadata['versions'][version]['start'], 0),
            length=extract_metadata['length'])
        lzd = lzma.LZMADecompressor()
        contents = lzd.decompress(contents)
        out_files = []

        #                                   salamand        salamandj       lifefrce        lifefrcej
        # 0x000000  0x20000     maincpu2                    X
        # 0x020000  0x40000     
        # 0x060000  0x08000     audiocpu    X               X
        # 0x068000  0x20000     krom        X               X               X               X
        # 0x088000  0x20000     maincpu2    X
        # 0x0A8000  0x40000     maincpu1    X               X               X
        # 0x0E8000  0x20000     maincpu2                                    X
        # 0x108000  0x40000     ???
        # 0x148000  0x08000     audiocpu                                    X
        # 0x150000  0x20000     ???
        # 0x170000  0x20000     maincpu2                                                    X
        # 0x170000  0x20000     maincpu1                                                    X
        # 0x1D0000  0x08000     audiocpu                                                    X
        # 0x1D8000  0x04000     vlm         X               X
        # 0x1DC000  0x04000     vlm                                         X               X

        # all common
        func_map = {}
        def krom(contents):
            out_chunks = {}
            out_chunks["587-c01.10a"] = transforms.cut(contents, 0x068000, length = 0x20000)
            return out_chunks
        func_map['krom'] = krom
        all_common_file_map = helpers.process_rom_files(contents, func_map)

        # salamand common
        func_map = {}
        def salamand_vlm(contents):
            out_chunks = {}
            out_chunks["587-d08.8g"] = transforms.cut(contents, 0x1D8000, length = 0x4000)
            return out_chunks
        func_map['vlm'] = salamand_vlm
        def salamand_audiocpu(contents):
            out_chunks = {}
            out_chunks["587-d09.11j"] = transforms.cut(contents, 0x060000, length = 0x8000)
            return out_chunks
        func_map['audiocpu'] = salamand_audiocpu
        salamand_common_file_map = helpers.process_rom_files(contents, func_map)

        # common maincpu1
        func_map = {}
        def common_maincpu1(contents):
            filenames = [
                "587-c03.17b",
                "587-c06.17c",
            ]
            contents = transforms.cut(contents, 0xA8000, length = 0x40000)
            chunks = transforms.deinterleave(contents, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['maincpu1'] = common_maincpu1
        common_maincpu1_file_map = helpers.process_rom_files(contents, func_map)

        # salamand
        func_map = {}
        def salamand_maincpu2(contents):
            filenames = [
                "587-d02.18b",
                "587-d05.18c",
            ]
            contents = transforms.cut(contents, 0x88000, length = 0x20000)
            chunks = transforms.deinterleave(contents, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['maincpu2'] = salamand_maincpu2
        func_map['all'] = helpers.existing_files_helper(all_common_file_map)
        func_map['salamand-common'] = helpers.existing_files_helper(salamand_common_file_map)
        func_map['common-maincpu1'] = helpers.existing_files_helper(common_maincpu1_file_map)
        out_files.append({'filename': 'salamand.zip', 'contents': helpers.build_rom(contents, func_map)})

        # salamandj
        func_map = {}
        def salamandj_maincpu2(contents):
            filenames = [
                "587-j02.18b",
                "587-j05.18c",
            ]
            contents = transforms.cut(contents, 0x00000, length = 0x20000)
            chunks = transforms.deinterleave(contents, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['maincpu2'] = salamandj_maincpu2
        func_map['all'] = helpers.existing_files_helper(all_common_file_map)
        func_map['salamand-common'] = helpers.existing_files_helper(salamand_common_file_map)
        func_map['common-maincpu1'] = helpers.existing_files_helper(common_maincpu1_file_map)
        out_files.append({'filename': 'salamandj.zip', 'contents': helpers.build_rom(contents, func_map)})

        # lifefrce
        func_map = {}
        def lifefrce_maincpu2(contents):
            filenames = [
                "587-k02.18b",
                "587-k05.18c",
            ]
            contents = transforms.cut(contents, 0xE8000, length = 0x20000)
            chunks = transforms.deinterleave(contents, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['maincpu2'] = lifefrce_maincpu2
        func_map['all'] = helpers.existing_files_helper(all_common_file_map)
        def lifefrce_vlm(contents):
            out_chunks = {}
            out_chunks["587-k08.8g"] = transforms.cut(contents, 0x1DC000, length = 0x4000)
            return out_chunks
        func_map['vlm'] = lifefrce_vlm
        def lifefrce_audiocpu(contents):
            out_chunks = {}
            out_chunks["587-k09.11j"] = transforms.cut(contents, 0x148000, length = 0x8000)
            return out_chunks
        func_map['audiocpu'] = lifefrce_audiocpu
        func_map['common-maincpu1'] = helpers.existing_files_helper(common_maincpu1_file_map)
        out_files.append({'filename': 'lifefrce.zip', 'contents': helpers.build_rom(contents, func_map)})

        # lifefrcej
        func_map = {}
        def lifefrcej_maincpu2(contents):
            filenames = [
                "587-n02.18b",
                "587-n05.18c",
            ]
            contents = transforms.cut(contents, 0x170000, length = 0x20000)
            chunks = transforms.deinterleave(contents, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['maincpu2'] = lifefrcej_maincpu2
        func_map['all'] = helpers.existing_files_helper(all_common_file_map)
        def lifefrcej_vlm(contents):
            out_chunks = {}
            out_chunks["587-n08.8g"] = transforms.cut(contents, 0x1DC000, length = 0x4000)
            return out_chunks
        func_map['vlm'] = lifefrcej_vlm
        def lifefrcej_audiocpu(contents):
            out_chunks = {}
            out_chunks["587-n09.11j"] = transforms.cut(contents, 0x1D0000, length = 0x8000)
            return out_chunks
        func_map['audiocpu'] = lifefrcej_audiocpu
        def lifefrcej_maincpu1(contents):
            filenames = [
                "587-n03.17b",
                "587-n06.17c",
            ]
            contents = transforms.cut(contents, 0x190000, length = 0x40000)
            chunks = transforms.deinterleave(contents, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['maincpu1'] = lifefrcej_maincpu1
        out_files.append({'filename': 'lifefrcej.zip', 'contents': helpers.build_rom(contents, func_map)})

        return out_files

    def _handle_twinbee(self, full_contents, version):
        out_files = []
        
        # Get FSE files Nemesis
        extract_metadata = self._get_extract_metadata('Nemesis')
        contents = transforms.cut(full_contents, 
            int(extract_metadata['versions'][version]['start'], 0),
            length=extract_metadata['length'])
        lzd = lzma.LZMADecompressor()
        contents = lzd.decompress(contents)

        func_map = {}
        def fse(contents):
            filenames = [
                "400-a01.fse",
                "400-a02.fse"
            ]
            chunks = []
            contents = transforms.cut(contents, 0x84000, length = 0x200)
            chunks = transforms.equal_split(contents, len(filenames))
            return dict(zip(filenames, chunks))
        func_map['fse'] = fse
        fse_file_map = helpers.process_rom_files(contents, func_map)

        extract_metadata = self._get_extract_metadata('Twinbee')
        # Get the maincpu1
        contents = transforms.cut(full_contents, 
            int(extract_metadata['versions'][version]['start_maincpu'], 0),
            length=extract_metadata['length'])
        lzd = lzma.LZMADecompressor()
        contents = lzd.decompress(contents)

        func_map = {}
        def maincpu1(contents):
            filenames = [
                "412-a07.12l",
                "412-a05.17l"
            ]
            chunks = transforms.deinterleave(contents, 2, 1)
            return dict(zip(filenames, chunks))
        func_map['maincpu1'] = maincpu1
        maincpu1_file_map = helpers.process_rom_files(contents, func_map)

        # Get the k rom
        contents = transforms.cut(full_contents,
            int(extract_metadata['versions'][version]['start_krom'], 0),
            length=extract_metadata['length'])
        lzd = lzma.LZMADecompressor()
        contents = lzd.decompress(contents)
        func_map = {}
        def krom(contents):
            filenames = [
                "400-a06.15l",
                "400-a04.10l",
                "400-e03.5l"
            ]
            maincpu2_contents = transforms.cut(contents, 0x0, length=65536)
            chunks = transforms.deinterleave(maincpu2_contents, 2, 1)
            chunks.append(transforms.cut(contents, 0x10400, length=8192))
            return dict(zip(filenames, chunks))
        func_map['k005289'] = krom
        krom_file_map = helpers.process_rom_files(contents, func_map)


        func_map = {}
        func_map['fse'] = helpers.existing_files_helper(fse_file_map)
        func_map['maincpu1'] = helpers.existing_files_helper(maincpu1_file_map)
        func_map['krom'] = helpers.existing_files_helper(krom_file_map)
        out_files.append({'filename': 'twinbee.zip', 'contents': helpers.build_rom(contents, func_map)})

        return out_files

    def lzma_multi_decomp(self, in_data, out_dir):
        out_data = b''

        offset = 0
        while True:
            offset = in_data.find(b']\x00\x00', offset)
            if offset == -1:
                print("out of input")
                break

            # Check LZMA header
            # 1 => 5d 00 00 01 00
            # 2 => 5d 00 00 10 00
            # 3 => 5d 00 00 08 00
            # 4 => 5d 00 00 10 00
            # 5 => 5d 00 00 20 00
            # 6 => 5d 00 00 40 00
            # 7 => 5d 00 00 80 00
            # 8 => 5d 00 00 00 01
            # 9 => 5d 00 00 00 02
            comp_mode = in_data[offset+3:offset+5]
            valid_modes = [
                b'\x01\x00',
                b'\x10\x00',
                b'\x08\x00',
                b'\x18\x00',
                b'\x20\x00',
                b'\x40\x00',
                b'\x80\x00',
                b'\x00\x01',
                b'\x00\x02'
            ]
            if comp_mode not in valid_modes:
                logger.debug(f'offset {hex(offset)}: skipping, invalid comp_type')
                offset += 1
            else:
                try:
                    target = in_data[offset:]
                    before_len = len(target)
                    magic_bytes = in_data[offset:offset+5]
                    lzd = lzma.LZMADecompressor()
                    curr_chunk = lzd.decompress(target)
                    after_len = len(lzd.unused_data)
                    consumed_bytes = before_len - after_len
                    logger.debug(f'offset {hex(offset)}: magic {magic_bytes}, consumed {consumed_bytes} bytes')
                    with open(os.path.join(out_dir, f'decompressed_blob_{hex(offset)}'), "wb") as out_file:
                        out_file.write(curr_chunk)
                    out_data += curr_chunk
                    offset += consumed_bytes
                except lzma.LZMAError:
                    logger.debug(f'offset {hex(offset)}: magic {magic_bytes}, invalid')
                    offset += 1
        return out_data
